# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from gi.repository import Adw
from gi.repository import Gtk

from gajim.common import app
from gajim.common.i18n import _
from gajim.common.modules.contacts import GroupchatContact

from gajim.gtk.builder import get_builder
from gajim.gtk.contact_name_entry import ContactNameEntry
from gajim.gtk.groupchat_affiliation import GroupchatAffiliation
from gajim.gtk.groupchat_blocks import GroupchatBlocks
from gajim.gtk.groupchat_config import GroupchatConfig
from gajim.gtk.groupchat_info import GroupChatInfoScrolled
from gajim.gtk.groupchat_manage import GroupchatManage
from gajim.gtk.groupchat_outcasts import GroupchatOutcasts
from gajim.gtk.groupchat_settings import GroupChatSettings
from gajim.gtk.omemo_trust_manager import OMEMOTrustManager
from gajim.gtk.sidebar_switcher import SideBarMenuItem
from gajim.gtk.sidebar_switcher import SideBarSwitcher
from gajim.gtk.structs import AccountJidParam
from gajim.gtk.window import GajimAppWindow


class GroupchatDetails(GajimAppWindow):
    def __init__(self, contact: GroupchatContact, page: str | None = None) -> None:
        GajimAppWindow.__init__(
            self,
            name="GroupchatDetails",
            title=_("Group Chat Details"),
            default_height=600,
        )

        self.account = contact.account
        self._client = app.get_client(contact.account)

        self._contact = contact
        self._contact.connect("avatar-update", self._on_avatar_update)
        self._contact.connect("disco-info-update", self._on_disco_info_update)

        self._ui = get_builder("groupchat_details.ui")

        self._switcher = SideBarSwitcher(width=250)
        self._switcher.set_with_menu(
            self._ui.main_stack,
            [
                SideBarMenuItem(
                    "information",
                    _("Information"),
                    group=_("Personal Settings"),
                    icon_name="lucide-user-symbolic",
                ),
                SideBarMenuItem(
                    "settings",
                    _("Settings"),
                    group=_("Personal Settings"),
                    icon_name="lucide-settings-symbolic",
                ),
                SideBarMenuItem(
                    "encryption-omemo",
                    _("Encryption (OMEMO)"),
                    group=_("Personal Settings"),
                    icon_name="lucide-lock-symbolic",
                ),
                SideBarMenuItem(
                    "blocks",
                    _("Blocked Participants"),
                    group=_("Personal Settings"),
                    icon_name="lucide-user-round-x-symbolic",
                ),
                SideBarMenuItem(
                    "manage",
                    _("Manage"),
                    group=_("Administration"),
                    icon_name="lucide-square-pen-symbolic",
                ),
                SideBarMenuItem(
                    "affiliations",
                    _("Affiliations"),
                    group=_("Administration"),
                    icon_name="lucide-users-symbolic",
                ),
                SideBarMenuItem(
                    "outcasts",
                    _("Outcasts"),
                    group=_("Administration"),
                    icon_name="lucide-users-symbolic",
                ),
                SideBarMenuItem(
                    "config",
                    _("Configuration"),
                    group=_("Administration"),
                    icon_name="lucide-settings-symbolic",
                ),
            ],
            visible=False,
        )

        toolbar = Adw.ToolbarView(content=self._switcher)
        toolbar.add_top_bar(Adw.HeaderBar())

        self._sidebar_page = Adw.NavigationPage(
            title=self._contact.name, tag="sidebar", child=toolbar
        )

        toolbar = Adw.ToolbarView(content=self._ui.main_stack)
        toolbar.add_top_bar(Adw.HeaderBar())

        content_page = Adw.NavigationPage(title=" ", tag="content", child=toolbar)

        nav = Adw.NavigationSplitView(sidebar=self._sidebar_page, content=content_page)

        self.set_child(nav)

        self._groupchat_manage: GroupchatManage | None = None

        self._add_groupchat_info()
        self._add_groupchat_settings()
        self._add_groupchat_encryption()
        self._add_blocks()

        if self._client.state.is_available and self._contact.is_joined:
            self._add_groupchat_manage()
            self._add_affiliations()
            self._add_outcasts()
            self._add_configuration()

        if page is not None:
            self._switcher.activate_item(page)
        else:
            self._switcher.activate_item("information")

    def _cleanup(self) -> None:
        self._switcher.run_destroy()
        del self._switcher
        del self._groupchat_manage
        del self._groupchat_info

    def _on_disco_info_update(
        self, _contact: GroupchatContact, _signal_name: str
    ) -> None:
        self._groupchat_info.set_info_from_contact(self._contact)

    def _on_avatar_update(self, _contact: GroupchatContact, _signal_name: str) -> None:
        assert self._groupchat_manage
        self._groupchat_manage.update_avatar()

    def _add_groupchat_manage(self) -> None:
        self._groupchat_manage = GroupchatManage(self.account, self._contact)
        self._ui.manage_box.append(self._groupchat_manage)
        self._switcher.set_item_visible("manage", True)

    def _add_groupchat_info(self) -> None:
        self._groupchat_info = GroupChatInfoScrolled(
            self._contact.account,
            width=600,
            edit_mode=True,
            show_users=False,
            add_padding=False,
        )
        self._connect(
            self._groupchat_info, "name-updated", self._on_contact_name_updated
        )
        self._groupchat_info.set_halign(Gtk.Align.FILL)
        self._groupchat_info.set_info_from_contact(self._contact)
        self._groupchat_info.set_subject(self._contact.subject)
        self._ui.info_container.append(self._groupchat_info)
        self._switcher.set_item_visible("information", True)

    def _add_groupchat_settings(self) -> None:
        self._ui.settings_box.append(GroupChatSettings(self.account, self._contact.jid))

        remove_history_button = Gtk.Button(label=_("Remove History…"))
        remove_history_button.set_halign(Gtk.Align.START)
        remove_history_button.add_css_class("destructive-action")
        params = AccountJidParam(account=self.account, jid=self._contact.jid)
        remove_history_button.set_action_target_value(params.to_variant())
        remove_history_button.set_action_name("app.remove-history")
        self._ui.settings_box.append(remove_history_button)
        self._switcher.set_item_visible("settings", True)

    def _add_groupchat_encryption(self) -> None:
        if self._contact.is_groupchat and self._contact.muc_context == "public":
            # OMEMO is not available for public group chats
            self._switcher.set_item_visible("encryption-omemo", False)
            return

        self._ui.encryption_box.set_child(
            OMEMOTrustManager(self._contact.account, self._contact)
        )
        self._switcher.set_item_visible("encryption-omemo", True)

    def _add_blocks(self) -> None:
        blocks = GroupchatBlocks(self._client, self._contact)
        self._ui.blocks_box.append(blocks)
        self._switcher.set_item_visible("blocks", True)

    def _add_affiliations(self) -> None:
        affiliations = GroupchatAffiliation(self._client, self._contact)
        self._ui.affiliation_box.append(affiliations)
        self._switcher.set_item_visible("affiliations", True)

    def _add_outcasts(self) -> None:
        affiliations = GroupchatOutcasts(self._client, self._contact)
        self._ui.outcasts_box.append(affiliations)
        self._switcher.set_item_visible("outcasts", True)

    def _add_configuration(self) -> None:
        config = GroupchatConfig(self._client, self._contact)
        self._ui.configuration_box.append(config)
        self._switcher.set_item_visible("config", True)

    def _on_contact_name_updated(self, _widget: ContactNameEntry, name: str) -> None:
        self._sidebar_page.set_title(name)



# Copyright (C) 2003-2014 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2005-2007 Nikos Kouremenos <kourem AT gmail.com>
# Copyright (C) 2006 Dimitur Kirov <dkirov AT gmail.com>
# Copyright (C) 2008-2009 Julien Pivotto <roidelapluie AT gmail.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import typing
from typing import Any

import logging

from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import GtkSource
from gi.repository import Pango

from gajim.common import app
from gajim.common import events
from gajim.common import ged
from gajim.common.ged import EventHelper
from gajim.common.i18n import _
from gajim.common.i18n import get_default_lang
from gajim.common.storage.archive import models as mod
from gajim.common.styling import PlainBlock
from gajim.common.styling import process
from gajim.common.types import ChatContactT

from gajim.gtk.alert import InformationAlertDialog
from gajim.gtk.completion.commands import CommandsCompletionProvider
from gajim.gtk.completion.emoji import EmojiCompletionProvider
from gajim.gtk.completion.nickname import NicknameCompletionProvider
from gajim.gtk.completion.popover import CompletionPopover
from gajim.gtk.const import MAX_MESSAGE_LENGTH
from gajim.gtk.menus import get_message_input_extra_context_menu
from gajim.gtk.util.misc import scroll_to
from gajim.gtk.widgets import GdkRectangle

if app.is_installed("SPELLING") or typing.TYPE_CHECKING:
    from gi.repository import Spelling

FORMAT_CHARS: dict[str, str] = {
    "bold": "*",
    "italic": "_",
    "strike": "~",
    "pre": "`",
}

log = logging.getLogger("gajim.gtk.message_input")


class MessageInputTextView(GtkSource.View, EventHelper):
    """
    A GtkSource.View for chat message input
    """

    __gsignals__ = {
        "buffer-changed": (
            GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.ACTION,
            None,
            (),
        ),
    }

    def __init__(self, parent: Gtk.Widget) -> None:
        GtkSource.View.__init__(
            self,
            accepts_tab=True,
            wrap_mode=Gtk.WrapMode.WORD_CHAR,
            margin_top=3,
            margin_bottom=3,
            valign=Gtk.Align.CENTER,
        )
        EventHelper.__init__(self)

        self._parent = parent
        self._completion_providers = [
            EmojiCompletionProvider(),
            CommandsCompletionProvider(),
            NicknameCompletionProvider(),
        ]

        self.add_css_class("gajim-conversation-text")
        self.add_css_class("message-input-textview")

        self._contact: ChatContactT | None = None

        self._text_buffer_manager = TextBufferManager(self)
        self._text_buffer_manager.connect("buffer-changed", self._on_buffer_changed)

        self._completion_popover = CompletionPopover(self)
        self._completion_popover.connect(
            "completion-picked", self._on_completion_picked
        )

        gesture_secondary_click = Gtk.GestureClick(button=Gdk.BUTTON_SECONDARY)
        gesture_secondary_click.connect("pressed", self._on_secondary_click)
        self.add_controller(gesture_secondary_click)

        focus_controller = Gtk.EventControllerFocus()
        focus_controller.connect("enter", self._on_focus_enter)
        focus_controller.connect("leave", self._on_focus_leave)
        self.add_controller(focus_controller)

        manager = app.app.get_shortcut_manager()
        manager.install_shortcuts(self, ["input", "input-global"])

        self.connect_after("paste-clipboard", self._after_paste_clipboard)

        self._speller_menu: Gio.MenuModel | None = None
        self._update_extra_menu()

        app.plugin_manager.gui_extension_point("message_input", self)

        self.register_events(
            [
                ("register-actions", ged.GUI2, self._on_register_actions),
            ]
        )

    def _on_register_actions(self, _event: events.RegisterActions) -> None:
        actions = [
            "input-clear",
            "input-focus",
            "input-bold",
            "input-italic",
            "input-strike",
            "input-emoji",
            "input-paste-as-quote",
            "input-paste-as-code-block",
            "input-mention",
        ]

        for action in actions:
            action = app.window.get_action(action)
            action.connect("activate", self._on_action)

    def _on_action(
        self, action: Gio.SimpleAction, param: GLib.Variant | None
    ) -> int | None:
        if self._contact is None:
            return

        action_name = action.get_name()
        log.info("Activate action: %s", action_name)

        match action_name:
            case "input-clear":
                self.clear()

            case "input-focus":
                self.grab_focus_delayed()

            case "input-bold" | "input-italic" | "input-strike":
                self._apply_formatting(action_name.removeprefix("input-"))

            case "input-emoji":
                self.emit("insert-emoji")

            case "input-paste-as-quote":
                self._paste_as_quote()

            case "input-paste-as-code-block":
                self._paste_as_code_block()

            case "input-mention":
                assert param is not None
                self._mention_participant(param.get_string())

            case _:
                pass

    def get_completion_popover(self) -> CompletionPopover:
        return self._completion_popover

    def grab_focus_delayed(self) -> bool:
        # Use this function if calling via GLib.idle_add to avoid
        # creating a loop raising CPU load
        self.grab_focus()
        return False

    def start_correction(self, message: mod.Message | None = None) -> None:
        if message is None:
            return

        self.clear()
        self.grab_focus()

        text = message.text
        assert text is not None

        self.insert_text(text)

    def end_correction(self) -> None:
        self.clear()

    def switch_contact(self, contact: ChatContactT) -> None:
        self._text_buffer_manager.switch_contact(contact)
        self.clear()
        self._contact = contact

    def set_speller_menu(self, menu: Gio.MenuModel) -> None:
        self._speller_menu = menu
        self._update_extra_menu()

    def _update_extra_menu(self) -> None:
        menu = get_message_input_extra_context_menu()
        if self._speller_menu is not None:
            menu.append_section(_("Spell Checking"), self._speller_menu)
        self.set_extra_menu(menu)

    def _on_buffer_changed(self, _text_buffer_manager: TextBufferManager) -> None:
        buf = self.get_buffer()
        if self._populate_completion(buf):
            _success, rect = self.compute_bounds(self._parent)
            point = rect.get_top_left()
            rect = GdkRectangle(int(point.x), int(point.y) - 20, 1, 1)
            self._completion_popover.set_pointing_to(rect)
            self._completion_popover.popup()
        else:
            self._completion_popover.popdown()

        self._on_text_changed()

        self.emit("buffer-changed")

    @staticmethod
    def _get_completion_candidate_bounds(
        text_buffer: Gtk.TextBuffer,
    ) -> tuple[Gtk.TextIter, Gtk.TextIter]:
        end = text_buffer.get_iter_at_mark(text_buffer.get_insert())

        _success, start_line_iter = text_buffer.get_iter_at_line(end.get_line())

        search_res = end.backward_search(
            " ",
            Gtk.TextSearchFlags.TEXT_ONLY | Gtk.TextSearchFlags.CASE_INSENSITIVE,
            start_line_iter,
        )
        if search_res is None:
            start = start_line_iter
        else:
            start, _ = search_res
            start.forward_char()
        return start, end

    def _populate_completion(self, buf: Gtk.TextBuffer) -> bool:
        start, end = self._get_completion_candidate_bounds(buf)
        candidate = buf.get_text(start, end, False)

        for provider in self._completion_providers:
            if not provider.check(candidate, start):
                continue

            if provider.populate(candidate, self._contact):
                self._completion_popover.set_provider(provider)
                return True
        return False

    def _on_completion_picked(
        self, popover: CompletionPopover, complete_string: str
    ) -> None:
        buf = self.get_buffer()
        start, end = self._get_completion_candidate_bounds(buf)
        buf.delete(start, end)
        buf.insert(start, complete_string)
        self.grab_focus()

    def _on_secondary_click(
        self,
        _gesture_click: Gtk.GestureClick,
        _n_press: int,
        x: float,
        y: float,
    ) -> None:
        # Place the cursor at click position to trigger an update
        # for spelling suggestions, see:
        # https://gitlab.gnome.org/GNOME/libspelling/-/issues/5
        buf = self.get_buffer()
        if buf.get_has_selection():
            # Don't place cursor if text is selected, otherwise the selection is changed
            return

        _, iter_, _ = self.get_iter_at_position(int(x), int(y))
        buf.place_cursor(iter_)
        return

    def _on_focus_enter(self, _focus_controller: Gtk.EventControllerFocus) -> None:
        scrolled = self.get_parent()
        assert scrolled
        scrolled.add_css_class("message-input-focus")

    def _on_focus_leave(self, _focus_controller: Gtk.EventControllerFocus) -> None:
        scrolled = self.get_parent()
        assert scrolled
        scrolled.remove_css_class("message-input-focus")
        if not self.has_focus():
            self._completion_popover.popdown()

    def _clear_tags(self) -> None:
        to_remove: list[Gtk.TextTag] = []

        def _check(tag: Gtk.TextTag) -> None:
            if tag.get_property("underline-rgba-set") is True:
                # Don’t remove spell checking underlines
                return
            to_remove.append(tag)

        buf = self.get_buffer()
        start, end = buf.get_bounds()
        tag_table = buf.get_tag_table()
        tag_table.foreach(_check)
        for tag in to_remove:
            buf.remove_tag(tag, start, end)

    def _on_text_changed(self) -> None:
        text = self.get_text()
        if not text:
            return

        self._clear_tags()

        if len(text) > MAX_MESSAGE_LENGTH:
            # Limit message styling processing
            return

        buf = self.get_buffer()
        result = process(text)
        for block in result.blocks:
            if isinstance(block, PlainBlock):
                for span in block.spans:
                    start_iter = buf.get_iter_at_offset(span.start + block.start)
                    end_iter = buf.get_iter_at_offset(span.end + block.start)
                    buf.apply_tag_by_name(span.name, start_iter, end_iter)

    def insert_text(self, text: str) -> None:
        self.get_buffer().insert_at_cursor(text)

    def insert_newline(self) -> None:
        # Reset IMContext to clear preedit state
        self.reset_im_context()
        buf = self.get_buffer()
        buf.insert_at_cursor("\n")
        mark = buf.get_insert()
        iter_ = buf.get_iter_at_mark(mark)
        if buf.get_end_iter().equal(iter_):
            GLib.idle_add(scroll_to, self.get_parent(), "bottom")

    @property
    def has_text(self) -> bool:
        buf = self.get_buffer()
        start, end = buf.get_bounds()
        text = buf.get_text(start, end, True)
        return text != ""

    def get_text(self) -> str:
        buf = self.get_buffer()
        start, end = buf.get_bounds()
        return self.get_buffer().get_text(start, end, True)

    @staticmethod
    def _after_paste_clipboard(textview: Gtk.TextView) -> None:
        buf = textview.get_buffer()
        mark = buf.get_insert()
        iter_ = buf.get_iter_at_mark(mark)
        if iter_.get_offset() == buf.get_end_iter().get_offset():
            GLib.idle_add(scroll_to, textview.get_parent(), "bottom")

    def _get_active_iters(self) -> tuple[Gtk.TextIter, Gtk.TextIter]:
        buf = self.get_buffer()
        return_val = buf.get_selection_bounds()
        if return_val:  # if something is selected
            start, end = return_val[0], return_val[1]
        else:
            start, end = buf.get_bounds()
        return (start, end)

    def _apply_formatting(self, formatting: str) -> None:
        format_char = FORMAT_CHARS[formatting]

        buf = self.get_buffer()
        start, end = self._get_active_iters()
        start_offset = start.get_offset()
        end_offset = end.get_offset()

        text = buf.get_text(start, end, True)
        if text.startswith(format_char) and text.endswith(format_char):
            # (Selected) text begins and ends with formatting chars
            # -> remove them
            buf.delete(start, buf.get_iter_at_offset(start_offset + 1))
            buf.delete(
                buf.get_iter_at_offset(end_offset - 2),
                buf.get_iter_at_offset(end_offset - 1),
            )
            return

        ext_start = buf.get_iter_at_offset(start_offset - 1)
        ext_end = buf.get_iter_at_offset(end_offset + 1)
        ext_text = buf.get_text(ext_start, ext_end, True)
        if ext_text.startswith(format_char) and ext_text.endswith(format_char):
            # (Selected) text is surrounded by formatting chars -> remove them
            buf.delete(ext_start, buf.get_iter_at_offset(start_offset))
            buf.delete(
                buf.get_iter_at_offset(end_offset - 1),
                buf.get_iter_at_offset(end_offset),
            )
            return

        # No formatting chars found at start/end or surrounding -> add them
        buf.insert(start, format_char, -1)
        buf.insert(buf.get_iter_at_offset(end_offset + 1), format_char, -1)
        buf.select_range(
            buf.get_iter_at_offset(start_offset), buf.get_iter_at_offset(end_offset + 2)
        )

    def clear(self, *args: Any) -> None:
        buf = self.get_buffer()
        buf.delete(*buf.get_bounds())

        # Action only available in Gtk 4.19
        # Use this as soon we set it as min Gtk
        # self.activate_action("text.clear")

    def undo(self, *args: Any) -> None:
        buf = self.get_buffer()
        if buf.get_can_undo():
            buf.undo()

    def redo(self, *args: Any) -> None:
        buf = self.get_buffer()
        if buf.get_can_redo():
            buf.redo()

    def _mention_participant(self, name: str) -> None:
        gc_refer_to_nick_char = app.settings.get("gc_refer_to_nick_char")
        text = f"{name}{gc_refer_to_nick_char} "
        self.insert_text(text)
        self.grab_focus()

    def insert_as_quote(self, text: str) -> None:
        text = "> " + text.replace("\n", "\n> ") + "\n"
        self.insert_text(text)
        self.grab_focus()

    def insert_as_code_block(self, text: str) -> None:
        self.insert_text(f"```\n{text}\n```")
        self.grab_focus()

    def _paste_as_quote(self) -> None:
        clipboard = self.get_clipboard()
        clipboard.read_text_async(
            None, self._on_clipboard_read_text_finished, "input-paste-as-quote"
        )

    def _paste_as_code_block(self) -> None:
        clipboard = self.get_clipboard()
        clipboard.read_text_async(
            None, self._on_clipboard_read_text_finished, "input-paste-as-code-block"
        )

    def _on_clipboard_read_text_finished(
        self, clipboard: Gdk.Clipboard, result: Gio.AsyncResult, action_name: str
    ) -> None:
        try:
            text = clipboard.read_text_finish(result)
        except Exception as e:
            InformationAlertDialog(_("Pasting Content Failed"), _("Error: %s") % e)
            return

        if text is None:
            log.info("No text pasted")
            return

        if action_name == "input-paste-as-quote":
            self.insert_as_quote(text)
        elif action_name == "input-paste-as-code-block":
            self.insert_as_code_block(text)


class TextBufferManager(GObject.Object):
    __gsignals__ = {
        "buffer-changed": (
            GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.ACTION,
            None,
            (),
        )
    }

    def __init__(self, message_input: MessageInputTextView) -> None:
        """A manager for GtkSource.Buffers. Having a Buffer per contact
        allows us to have a GtkSource.UndoManager per contact.
        """
        super().__init__()

        self._message_input = message_input
        self._contact: ChatContactT | None = None

        self._text_buffers: dict[ChatContactT, GtkSource.Buffer] = {}
        self._text_buffer_handlers: dict[ChatContactT, int] = {}

        self._spelling_adapters: dict[ChatContactT, Spelling.TextBufferAdapter] = {}
        self._spelling_language_handlers: dict[ChatContactT, int] = {}

        app.settings.connect_signal("use_speller", self._on_toggle_spell_check)

    def switch_contact(self, contact: ChatContactT) -> None:
        buf = self._text_buffers.get(contact)
        if buf is None:
            buf = GtkSource.Buffer(
                implicit_trailing_newline=False,
                highlight_matching_brackets=False,
            )
            buf.create_tag("strong", weight=Pango.Weight.BOLD)
            buf.create_tag("emphasis", style=Pango.Style.ITALIC)
            buf.create_tag("strike", strikethrough=True)
            buf.create_tag("pre", family="monospace")

            style_scheme = self._get_style_scheme(buf)
            if style_scheme is not None:
                buf.set_style_scheme(style_scheme)

            self._text_buffers[contact] = buf

            self._init_spell_checker(contact)
        else:
            buffer_handler = self._text_buffer_handlers[contact]
            buf.disconnect(buffer_handler)

            self._disconnect_spell_checker(contact)

        # Since the buffer changes when switching contacts, MessageActionsBox
        # cannot connect to a buffer's 'changed' signal.
        # Instead, we (re)connect each buffer and relay its 'changed' signal
        # via a custom 'buffer-changed' signal.
        self._text_buffer_handlers[contact] = buf.connect(
            "changed", self._on_buffer_changed
        )
        self._message_input.set_buffer(buf)

        self._connect_spell_checker(contact)

        self._contact = contact

    def _on_buffer_changed(self, _text_buffer: GtkSource.Buffer) -> None:
        self.emit("buffer-changed")

    def _get_style_scheme(self, buf: GtkSource.Buffer) -> GtkSource.StyleScheme | None:
        style_scheme_manager = GtkSource.StyleSchemeManager.get_default()
        if app.css_config.prefer_dark:
            return style_scheme_manager.get_scheme("Adwaita-dark")
        return style_scheme_manager.get_scheme("Adwaita")

    def _init_spell_checker(self, contact: ChatContactT) -> None:
        if not app.is_installed("SPELLING"):
            return

        provider = Spelling.Provider.get_default()
        checker = Spelling.Checker.new(
            provider, self._get_spell_checker_language(contact)
        )

        text_buffer = self._text_buffers[contact]
        self._spelling_adapters[contact] = Spelling.TextBufferAdapter.new(
            text_buffer,
            checker,
        )

    def _connect_spell_checker(self, contact: ChatContactT) -> None:
        if not app.is_installed("SPELLING"):
            return

        adapter = self._spelling_adapters[contact]
        speller_menu = adapter.get_menu_model()

        self._message_input.set_speller_menu(speller_menu)
        self._message_input.insert_action_group("spelling", adapter)

        adapter.set_enabled(app.settings.get("use_speller"))

        checker = adapter.get_checker()
        assert checker is not None
        self._spelling_language_handlers[contact] = checker.connect(
            "notify::language", self._on_language_changed
        )

    def _disconnect_spell_checker(self, contact: ChatContactT) -> None:
        if not app.is_installed("SPELLING"):
            return

        adapter = self._spelling_adapters[contact]
        checker = adapter.get_checker()
        assert checker is not None
        checker_handler_id = self._spelling_language_handlers[contact]
        checker.disconnect(checker_handler_id)

        self._message_input.insert_action_group("spelling", None)

    def _on_toggle_spell_check(self, *args: Any) -> None:
        if not app.is_installed("SPELLING"):
            return

        use_spell_check = app.settings.get("use_speller")
        for adapter in self._spelling_adapters.values():
            adapter.set_enabled(use_spell_check)

    def _get_spell_checker_language(self, contact: ChatContactT) -> str:
        lang = contact.settings.get("speller_language")
        if not lang:
            # use the default one
            lang = app.settings.get("speller_language")
            if not lang:
                lang = get_default_lang()

        return lang or "en"

    def _on_language_changed(self, checker: Spelling.Checker, _param: Any) -> None:
        language_code = checker.get_language()
        if language_code is not None:
            assert self._contact is not None
            handler_id = self._spelling_language_handlers[self._contact]
            with checker.handler_block(handler_id):
                self._contact.settings.set("speller_language", language_code)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast
from typing import Literal
from typing import overload

import json
import logging

from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from nbxmpp.client import Client as NBXMPPClient
from nbxmpp.const import ConnectionProtocol
from nbxmpp.const import ConnectionType
from nbxmpp.const import Mode
from nbxmpp.const import StreamError
from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import RegisterStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.modules.dataforms import SimpleDataForm
from nbxmpp.protocol import JID
from nbxmpp.protocol import validate_domainpart
from nbxmpp.structs import ProxyData
from nbxmpp.structs import RegisterData
from nbxmpp.task import Task

from gajim.common import app
from gajim.common.const import GIO_TLS_ERRORS
from gajim.common.const import SASL_ERRORS
from gajim.common.events import StanzaReceived
from gajim.common.events import StanzaSent
from gajim.common.file_transfer_manager import FileTransfer
from gajim.common.helpers import get_global_proxy
from gajim.common.helpers import get_proxy
from gajim.common.i18n import _
from gajim.common.util.jid import validate_jid
from gajim.common.util.text import get_country_flag_from_code

from gajim.gtk.assistant import Assistant
from gajim.gtk.assistant import AssistantErrorPage
from gajim.gtk.assistant import AssistantPage
from gajim.gtk.assistant import AssistantProgressPage
from gajim.gtk.assistant import AssistantSuccessPage
from gajim.gtk.dataform import DataFormWidget
from gajim.gtk.dropdown import GajimDropDown
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import clear_listbox
from gajim.gtk.util.misc import container_remove_all
from gajim.gtk.util.misc import get_ui_string
from gajim.gtk.util.misc import open_uri
from gajim.gtk.util.styling import get_color_for_account
from gajim.gtk.util.window import get_app_window
from gajim.gtk.util.window import open_window

CustomHostT = tuple[str, ConnectionProtocol, ConnectionType]

log = logging.getLogger("gajim.gtk.account_wizard")


class AccountWizard(Assistant):
    def __init__(self) -> None:
        Assistant.__init__(self, name="AccountWizard", height=500)
        self._client: NBXMPPClient | None = None
        self._method: Literal["login"] | Literal["signup"] = "login"
        self._destroyed: bool = False

        self.add_button("back", _("Back"))
        self.add_button(
            "signup", _("Sign Up"), complete=True, css_class="suggested-action"
        )
        self.add_button("connect", _("Connect"), css_class="suggested-action")
        self.add_button("next", _("Next"), css_class="suggested-action")
        self.add_button(
            "login", _("Log In"), complete=True, css_class="suggested-action"
        )

        self.add_pages(
            {
                "login": WizardLoginPage(),
                "signup": WizardSignupPage(),
                "advanced": WizardAdvancedPage(),
                "security-warning": WizardSecurityPage(),
                "form": WizardFormPage(),
                "redirect": WizardRedirectPage(),
                "success": WizardSuccessPage(),
                "error": WizardErrorPage(),
            }
        )

        self.add_default_page("progress")

        login_page = self.get_page("login")
        self._connect(login_page, "clicked", self._on_button_clicked)
        self._connect(self, "button-clicked", self._on_assistant_button_clicked)
        self._connect(self, "page-changed", self._on_page_changed)

        self.show_first_page()

    @overload
    def get_page(self, name: Literal["login"]) -> WizardLoginPage: ...

    @overload
    def get_page(self, name: Literal["signup"]) -> WizardSignupPage: ...

    @overload
    def get_page(self, name: Literal["advanced"]) -> WizardAdvancedPage: ...

    @overload
    def get_page(self, name: Literal["security-warning"]) -> WizardSecurityPage: ...

    @overload
    def get_page(self, name: Literal["form"]) -> WizardFormPage: ...

    @overload
    def get_page(self, name: Literal["redirect"]) -> WizardRedirectPage: ...

    @overload
    def get_page(self, name: Literal["success"]) -> WizardSuccessPage: ...

    @overload
    def get_page(self, name: Literal["error"]) -> WizardErrorPage: ...

    @overload
    def get_page(self, name: Literal["progress"]) -> AssistantProgressPage: ...

    def get_page(self, name: str) -> AssistantPage:
        return self._pages[name]

    def get_current_method(self) -> Literal["login"] | Literal["signup"]:
        return self._method

    def _on_button_clicked(self, _page: Gtk.Widget, button_name: str) -> None:
        if button_name == "login":
            if self.get_page("login").is_advanced():
                self.show_page("advanced", Gtk.StackTransitionType.SLIDE_LEFT)
            else:
                self._test_credentials()

        elif button_name == "signup":
            self.get_page("signup").update_providers_list()
            self.show_page("signup", Gtk.StackTransitionType.SLIDE_LEFT)

    def _on_assistant_button_clicked(
        self, _assistant: Assistant, button_name: str
    ) -> None:
        page = self.get_current_page()

        if button_name == "login":
            if page == "advanced":
                self._test_credentials()

            elif page == "security-warning":
                if self.get_page("security-warning").trust_certificate:
                    cert = self.get_page("security-warning").cert
                    assert cert is not None
                    app.cert_store.add_certificate(cert)
                self._test_credentials(ignore_all_errors=True)

        elif button_name == "signup":
            if self.get_page("signup").is_anonymous():
                domain = self.get_page("signup").get_server()

                if app.settings.account_exists(f"anon@{domain}"):
                    self._show_error_page(
                        _("Account exists already"),
                        _("Account exists already"),
                        _("This account has already been added"),
                    )
                    return

            if page == "signup":
                if self.get_page("signup").is_advanced():
                    self.show_page("advanced", Gtk.StackTransitionType.SLIDE_LEFT)

                elif self.get_page("signup").is_anonymous():
                    self._test_anonymous_server()

                else:
                    self._register_with_server()

            elif page == "advanced":
                if self.get_page("signup").is_anonymous():
                    self._test_anonymous_server()
                else:
                    self._register_with_server()

            elif page == "security-warning":
                if self.get_page("security-warning").trust_certificate:
                    cert = self.get_page("security-warning").cert
                    assert cert is not None
                    app.cert_store.add_certificate(cert)

                if self.get_page("signup").is_anonymous():
                    self._test_anonymous_server(ignore_all_errors=True)

                else:
                    self._register_with_server(ignore_all_errors=True)

            elif page == "form":
                self._show_progress_page(
                    _("Creating Account..."), _("Trying to create account...")
                )
                self._submit_form()

        elif button_name == "connect":
            if page == "success":
                account = self.get_page("success").account
                assert account is not None
                app.app.enable_account(account)
                self.close()

        elif button_name == "back":
            if page == "signup":
                self.show_page("login", Gtk.StackTransitionType.SLIDE_RIGHT)

            elif page in ("advanced", "error", "security-warning"):
                if (
                    page == "error"
                    and self._method == "signup"
                    and self.get_page("form").has_form
                ):
                    self.show_page("form", Gtk.StackTransitionType.SLIDE_RIGHT)
                else:
                    self.show_page(self._method, Gtk.StackTransitionType.SLIDE_RIGHT)

            elif page == "form":
                self.show_page("signup", Gtk.StackTransitionType.SLIDE_RIGHT)
                self.get_page("form").remove_form()
                self._disconnect()

            elif page == "redirect":
                self.show_page("login", Gtk.StackTransitionType.SLIDE_RIGHT)

    def _on_page_changed(self, _assistant: Assistant, page_name: str) -> None:
        if page_name == "signup":
            self._method = page_name
            self.get_page("signup").focus()

        elif page_name == "login":
            self._method = page_name
            self.get_page("login").focus()

        elif page_name == "form":
            self.get_page("form").focus()

    def update_proxy_list(self) -> None:
        self.get_page("advanced").update_proxy_list()

    def _get_proxy_data(self, advanced: bool) -> ProxyData | None:
        if advanced:
            proxy_name = self.get_page("advanced").get_proxy()
            proxy_data = get_proxy(proxy_name)
            if proxy_data is not None:
                return proxy_data

        return get_global_proxy()

    def _get_base_client(
        self, address: JID, mode: Mode, advanced: bool, ignore_all_errors: bool
    ) -> NBXMPPClient:
        client = NBXMPPClient(log_context="Account Wizard")
        client.set_domain(address.domain)
        client.set_username(address.localpart)
        client.set_mode(mode)
        client.set_ignore_tls_errors(ignore_all_errors)
        client.set_accepted_certificates(app.cert_store.get_certificates())

        if advanced:
            custom_host = self.get_page("advanced").get_custom_host()
            if custom_host is not None:
                client.set_custom_host(*custom_host)

        proxy_data = self._get_proxy_data(advanced)
        client.set_proxy(proxy_data)
        client.subscribe("disconnected", self._on_disconnected)
        client.subscribe("connection-failed", self._on_connection_failed)
        client.subscribe("stanza-sent", self._on_stanza_sent)
        client.subscribe("stanza-received", self._on_stanza_received)
        return client

    def _disconnect(self) -> None:
        if self._client is None:
            return
        self._client.remove_subscriptions()
        self._client.disconnect()
        self._client = None

    @staticmethod
    def _on_stanza_sent(_client: NBXMPPClient, _signal_name: str, stanza: Any) -> None:
        app.ged.raise_event(StanzaSent(account="AccountWizard", stanza=stanza))

    @staticmethod
    def _on_stanza_received(
        _client: NBXMPPClient, _signal_name: str, stanza: Any
    ) -> None:
        app.ged.raise_event(StanzaReceived(account="AccountWizard", stanza=stanza))

    def _request_host_meta_and_connect(
        self, client: NBXMPPClient, advanced: bool
    ) -> None:
        def _on_host_meta_response(obj: FileTransfer) -> None:
            if self._destroyed:
                return

            try:
                result = obj.get_result()
            except Exception as error:
                log.warning("Error while requesting host-meta data: %s", error)
            else:
                log.info("Received host meta data with length: %s", len(result.content))
                client.set_host_meta_data(result.content)

            client.connect()

        app.ftm.http_request(
            "GET",
            f"https://{client.domain}/.well-known/host-meta",
            proxy=self._get_proxy_data(advanced),
            callback=_on_host_meta_response,
        )

    def _test_credentials(self, ignore_all_errors: bool = False) -> None:
        self._show_progress_page(_("Connecting..."), _("Connecting to server..."))
        jid, password = self.get_page("login").get_credentials()
        address = JID.from_string(jid)
        advanced = self.get_page("login").is_advanced()

        self._client = self._get_base_client(
            address, Mode.LOGIN_TEST, advanced, ignore_all_errors
        )

        self._client.set_password(password)
        self._client.subscribe("login-successful", self._on_login_successful)
        self._request_host_meta_and_connect(self._client, advanced)

    def _test_anonymous_server(self, ignore_all_errors: bool = False) -> None:
        self._show_progress_page(_("Connecting..."), _("Connecting to server..."))
        domain = self.get_page("signup").get_server()
        advanced = self.get_page("signup").is_advanced()

        address = JID(localpart=None, domain=domain, resource=None)

        self._client = self._get_base_client(
            address, Mode.ANONYMOUS_TEST, advanced, ignore_all_errors
        )

        self._client.subscribe("anonymous-supported", self._on_anonymous_supported)
        self._request_host_meta_and_connect(self._client, advanced)

    def _register_with_server(self, ignore_all_errors: bool = False) -> None:
        self._show_progress_page(_("Connecting..."), _("Connecting to server..."))
        domain = self.get_page("signup").get_server()
        advanced = self.get_page("signup").is_advanced()

        address = JID(localpart=None, domain=domain, resource=None)

        self._client = self._get_base_client(
            address, Mode.REGISTER, advanced, ignore_all_errors
        )

        self._client.subscribe("connected", self._on_connected)
        self._request_host_meta_and_connect(self._client, advanced)

    def _on_login_successful(self, client: NBXMPPClient, _signal_name: str) -> None:
        assert client.domain is not None
        assert client.password is not None
        account = self._generate_account_name(client.domain)
        proxy_name = None
        if client.proxy is not None:
            proxy_name = self.get_page("advanced").get_proxy()

        address = JID(localpart=client.username, domain=client.domain, resource=None)

        app.app.create_account(
            account, address, client.password, proxy_name, client.custom_host
        )
        self.get_page("success").set_account(account)
        self.show_page("success", Gtk.StackTransitionType.SLIDE_LEFT)
        self._disconnect()

    def _on_connected(self, client: NBXMPPClient, _signal_name: str) -> None:
        client.get_module("Register").request_register_form(
            callback=self._on_register_form
        )

    def _on_anonymous_supported(self, client: NBXMPPClient, _signal_name: str) -> None:
        assert client.domain is not None
        assert client.password is not None
        account = self._generate_account_name(client.domain)
        proxy_name = None
        if client.proxy is not None:
            proxy_name = self.get_page("advanced").get_proxy()

        address = JID(localpart=None, domain=client.domain, resource=None)

        app.app.create_account(
            account,
            address,
            client.password,
            proxy_name,
            client.custom_host,
            anonymous=True,
        )

        self.get_page("success").set_account(account)
        self.show_page("success", Gtk.StackTransitionType.SLIDE_LEFT)
        self._disconnect()

    def _on_disconnected(self, client: NBXMPPClient, _signal_name: str) -> None:
        domain, error, text = client.get_error()
        if domain == StreamError.SASL:
            assert error is not None
            if error == "anonymous-not-supported":
                self._show_error_page(
                    _("Anonymous login not supported"),
                    _("Anonymous login not supported"),
                    _("This server does not support anonymous login."),
                )
            else:
                self._show_error_page(
                    _("Authentication failed"), SASL_ERRORS.get(error, ""), text or ""
                )

        elif domain == StreamError.BAD_CERTIFICATE:
            assert self._client is not None
            cert, errors = self._client.peer_certificate
            assert cert is not None
            assert errors is not None
            self.get_page("security-warning").set_warning(cert, list(errors))
            self.show_page("security-warning", Gtk.StackTransitionType.SLIDE_LEFT)

        elif domain == StreamError.REGISTER:
            if error == "register-not-supported":
                self._show_error_page(
                    _("Signup not allowed"),
                    _("Signup not allowed"),
                    _("This server does not allow signup."),
                )

        else:
            self._show_error_page(_("Error"), _("Error"), text or error or "")

        self.get_page("form").remove_form()
        assert self._client is not None
        self._client.destroy()
        self._client = None

    def _on_connection_failed(self, _client: NBXMPPClient, _signal_name: str) -> None:
        self._show_error_page(
            _("Connection failed"),
            _("Connection failed"),
            _(
                "Gajim was not able to reach the server. "
                "Make sure your XMPP address is correct."
            ),
        )
        assert self._client is not None
        self._client.destroy()
        self._client = None

    def _show_error_page(self, title: str, heading: str, text: str) -> None:
        self.get_page("error").set_title(title)
        self.get_page("error").set_heading(heading)
        self.get_page("error").set_text(text or "")
        self.show_page("error", Gtk.StackTransitionType.SLIDE_LEFT)

    def _show_progress_page(self, title: str, text: str) -> None:
        self.get_page("progress").set_title(title)
        self.get_page("progress").set_text(text)
        self.show_page("progress", Gtk.StackTransitionType.SLIDE_LEFT)

    @staticmethod
    def _generate_account_name(domain: str) -> str:
        i = 1
        while domain in app.settings.get_accounts():
            domain = domain + str(i)
            i += 1
        return domain

    def _on_register_form(self, task: Task) -> None:
        try:
            result = cast(RegisterData, task.finish())
        except (StanzaError, MalformedStanzaError) as error:
            self._show_error_page(_("Error"), _("Error"), error.get_text())
            self._disconnect()
            return

        if result.bob_data is not None:
            algo_hash = result.bob_data.cid.split("@")[0]
            app.bob_cache[algo_hash] = result.bob_data.data

        form = result.form
        if result.form is None:
            form = result.fields_form

        if form is not None:
            self.get_page("form").add_form(form)

        elif result.oob_url is not None:
            self.get_page("redirect").set_redirect(result.oob_url, result.instructions)
            self.show_page("redirect", Gtk.StackTransitionType.SLIDE_LEFT)
            self._disconnect()
            return

        self.show_page("form", Gtk.StackTransitionType.SLIDE_LEFT)

    def _submit_form(self) -> None:
        self.get_page("progress").set_text(_("Account is being created"))
        self.show_page("progress", Gtk.StackTransitionType.SLIDE_LEFT)

        form = self.get_page("form").get_submit_form()
        assert self._client is not None
        self._client.get_module("Register").submit_register_form(
            form, callback=self._on_register_result
        )

    def _on_register_result(self, task: Task) -> None:
        try:
            task.finish()
        except RegisterStanzaError as error:
            self._set_error_text(error)
            if error.type != "modify":
                self.get_page("form").remove_form()
                self._disconnect()
                return

            register_data = error.get_data()
            form = register_data.form
            if register_data.form is None:
                form = register_data.fields_form

            if form is not None:
                self.get_page("form").add_form(form)

            else:
                self.get_page("form").remove_form()
                self._disconnect()
            return

        except (StanzaError, MalformedStanzaError) as error:
            self._set_error_text(error)
            self.get_page("form").remove_form()
            self._disconnect()
            return

        username, password = self.get_page("form").get_credentials()
        assert self._client is not None
        assert self._client.domain is not None
        account = self._generate_account_name(self._client.domain)

        proxy_name = None
        if self._client.proxy is not None:
            proxy_name = self.get_page("advanced").get_proxy()

        address = JID(localpart=username, domain=self._client.domain, resource=None)

        app.app.create_account(
            account, address, password, proxy_name, self._client.custom_host
        )

        self.get_page("success").set_account(account)
        self.show_page("success", Gtk.StackTransitionType.SLIDE_LEFT)
        self.get_page("form").remove_form()
        self._disconnect()

    def _set_error_text(
        self, error: StanzaError | RegisterStanzaError | MalformedStanzaError
    ) -> None:
        error_text = error.get_text()
        if not error_text:
            error_text = _(
                "The server rejected the registration without an error message"
            )
        self._show_error_page(_("Error"), _("Error"), error_text)

    def _cleanup(self, *args: Any) -> None:
        Assistant._cleanup(self)
        self._disconnect()
        self._destroyed = True


@Gtk.Template(string=get_ui_string("wizard/login_page.ui"))
class WizardLoginPage(AssistantPage):
    __gtype_name__ = "WizardLoginPage"
    __gsignals__ = {
        "clicked": (GObject.SignalFlags.RUN_LAST, None, (str,)),
    }

    _log_in_address_entry: Gtk.Entry = Gtk.Template.Child()
    _log_in_password_entry: Gtk.Entry = Gtk.Template.Child()
    _login_advanced_checkbutton: Gtk.CheckButton = Gtk.Template.Child()
    _log_in_button: Gtk.Button = Gtk.Template.Child()
    _sign_up_button: Gtk.Button = Gtk.Template.Child()

    def __init__(self) -> None:
        AssistantPage.__init__(self)
        self.title = _("Add Account")

        self._connect(
            self._log_in_address_entry, "activate", self._on_address_entry_activate
        )
        self._connect(self._log_in_address_entry, "changed", self._on_address_changed)
        self._connect(self._log_in_password_entry, "changed", self._set_complete)
        self._connect(
            self._log_in_password_entry, "activate", self._on_password_entry_activate
        )
        self._connect(self._log_in_button, "clicked", self._on_login)
        self._connect(self._sign_up_button, "clicked", self._on_signup)

    def focus(self) -> None:
        self._log_in_address_entry.grab_focus()

    def _on_login(self, *args: Any) -> None:
        self.emit("clicked", "login")

    def _on_signup(self, *args: Any) -> None:
        self.emit("clicked", "signup")

    def _on_address_changed(self, entry: Gtk.Entry) -> None:
        self._set_complete()

    def _show_icon(self, show: bool) -> None:
        icon = "lucide-circle-alert-symbolic" if show else None
        self._log_in_address_entry.set_icon_from_icon_name(
            Gtk.EntryIconPosition.SECONDARY, icon
        )

    def _on_address_entry_activate(self, _entry: Gtk.Entry) -> None:
        GLib.idle_add(self._grab_password_entry_focus)

    def _grab_password_entry_focus(self) -> bool:
        self._log_in_password_entry.grab_focus()
        return False

    def _on_password_entry_activate(self, _entry: Gtk.Entry) -> None:
        if self._log_in_button.get_sensitive():
            self._log_in_button.activate()

    def _validate_jid(self, address: str) -> bool:
        if not address:
            self._show_icon(False)
            return False

        try:
            jid = validate_jid(address, type_="bare")
            if jid.resource:
                raise ValueError
        except ValueError:
            self._show_icon(True)
            self._log_in_address_entry.set_icon_tooltip_text(
                Gtk.EntryIconPosition.SECONDARY, _("Invalid Address")
            )
            return False

        if app.settings.account_exists(address):
            self._show_icon(True)
            self._log_in_address_entry.set_icon_tooltip_text(
                Gtk.EntryIconPosition.SECONDARY,
                _("This account has already been added"),
            )
            return False

        self._show_icon(False)
        return True

    def _set_complete(self, *args: Any) -> None:
        address = self._validate_jid(self._log_in_address_entry.get_text())
        password = self._log_in_password_entry.get_text()
        self._log_in_button.set_sensitive(bool(address and password))

    def is_advanced(self) -> bool:
        return self._login_advanced_checkbutton.get_active()

    def get_credentials(self) -> tuple[str, str]:
        return (
            self._log_in_address_entry.get_text(),
            self._log_in_password_entry.get_text(),
        )


@Gtk.Template(string=get_ui_string("wizard/signup_page.ui"))
class WizardSignupPage(AssistantPage):
    __gtype_name__ = "WizardSignupPage"

    _sign_up_advanced_checkbutton: Gtk.CheckButton = Gtk.Template.Child()
    _sign_up_anonymously: Gtk.CheckButton = Gtk.Template.Child()
    _sign_up_info_grid: Gtk.Grid = Gtk.Template.Child()
    _server_comboboxtext_sign_up: Gtk.ComboBoxText = Gtk.Template.Child()
    _update_provider_list_icon: Gtk.Image = Gtk.Template.Child()
    _popover: WizardServerRecommendationPopover = Gtk.Template.Child()

    def __init__(self) -> None:
        AssistantPage.__init__(self)
        self.complete: bool = False
        self.title: str = _("Create New Account")

        self._servers: list[dict[str, Any]] = []
        self._provider_list_request: FileTransfer | None = None

        entry = self._server_comboboxtext_sign_up.get_child()
        assert isinstance(entry, Gtk.Entry)
        self._entry = entry
        self._entry.set_activates_default(True)
        self._entry.set_placeholder_text("example.org")
        self._entry.set_input_purpose(Gtk.InputPurpose.URL)

        self._entry_completion = Gtk.EntryCompletion(
            text_column=0,
            inline_completion=True,
            popup_completion=False,
            popup_single_match=False,
        )
        self._entry.set_completion(self._entry_completion)

        self._connect(self._entry, "changed", self._set_complete)

    def do_unroot(self) -> None:
        AssistantPage.do_unroot(self)
        if self._provider_list_request is not None:
            self._provider_list_request.cancel()
            self._provider_list_request = None

    def focus(self) -> None:
        self._entry.grab_focus()

    def update_providers_list(self) -> None:
        if len(self._servers) > 0 or self._provider_list_request is not None:
            # A request has already been started
            return

        self._server_comboboxtext_sign_up.set_sensitive(False)
        self._update_provider_list_icon.add_css_class("spin")

        self._provider_list_request = app.ftm.http_request(
            "GET",
            app.settings.get_app_setting("providers_list_url"),
            callback=self._on_download_provider_list_finished,
        )

    def _on_download_provider_list_finished(self, obj: FileTransfer) -> None:
        self._provider_list_request = None
        self._server_comboboxtext_sign_up.set_sensitive(True)
        self._update_provider_list_icon.remove_css_class("spin")

        try:
            result = obj.get_result()
        except Exception as error:
            log.warning("Error while downloading provider list: %s", error)
            self._update_provider_list_icon.set_from_icon_name(
                "lucide-circle-x-symbolic"
            )
            self._update_provider_list_icon.set_tooltip_text(
                _("Could not update providers list")
            )
            return

        self._update_provider_list_icon.set_from_icon_name("lucide-check-symbolic")
        self._update_provider_list_icon.set_tooltip_text(
            _("Providers list is up to date")
        )

        self._servers = json.loads(result.content)
        self._create_server_completion()

    def _create_server_completion(self) -> None:
        servers_model = Gtk.ListStore(str, str)

        for server in self._servers:
            server_locations = " ".join(
                get_country_flag_from_code(location)
                for location in server["serverLocations"]
            )
            servers_model.append((server["jid"], server_locations))

        self._entry_completion.set_model(servers_model)

        # Sign up combobox and entry
        self._server_comboboxtext_sign_up.set_model(servers_model)
        cell_area = self._server_comboboxtext_sign_up.get_area()
        assert cell_area is not None

        language_renderer = Gtk.CellRendererText(xalign=1)
        cell_area.add(language_renderer)
        cell_area.attribute_connect(language_renderer, "text", 1)

        self._connect(
            self._server_comboboxtext_sign_up,
            "changed",
            self._on_sign_up_server_changed,
        )

    def _on_sign_up_server_changed(self, combo: Gtk.ComboBox) -> None:
        if len(self._servers) == 0:
            return

        selection = combo.get_active()
        server_data = self._servers[selection]
        if server_data["jid"] != self._entry.get_text():
            # Entry was changed manually
            self._clear_server_info_box()
            return

        self._show_server_infos(server_data)

    def _show_server_infos(self, server_data: dict[str, Any]) -> None:
        self._clear_server_info_box()

        heading = Gtk.Label(label=_("Provider Infos"), halign=Gtk.Align.START)
        heading.add_css_class("bold")
        self._sign_up_info_grid.attach(heading, 0, 0, 2, 1)

        provider_category = server_data["category"]
        image = Gtk.Image.new_from_icon_name("lucide-check-symbolic")
        label = Gtk.Label(
            label=_("Provider Category: %s") % provider_category,
            halign=Gtk.Align.START,
            hexpand=True,
        )
        self._sign_up_info_grid.attach(image, 0, 1, 1, 1)
        self._sign_up_info_grid.attach(label, 1, 1, 1, 1)

        providers_link = Gtk.LinkButton(
            label=_("More infos at providers.xmpp.net"),
            uri=f"https://providers.xmpp.net/provider/{server_data['jid']}/",
            halign=Gtk.Align.START,
        )
        self._sign_up_info_grid.attach(providers_link, 0, 2, 2, 1)

        self._sign_up_info_grid.attach(Gtk.Separator(), 0, 3, 2, 1)

    def _clear_server_info_box(self) -> None:
        container_remove_all(self._sign_up_info_grid)

    def _on_visit_server(self, _button: Gtk.Button) -> int:
        server = self._entry.get_text().strip()
        server = f"https://{server}"
        open_uri(server)
        return Gdk.EVENT_STOP

    def _set_complete(self, *args: Any) -> None:
        try:
            server = self.get_server()
        except Exception:
            self.complete = False
            self._popover.set_website(None)
        else:
            self.complete = True
            self._popover.set_website(server)

        self.update_page_complete()

    def is_anonymous(self) -> bool:
        return self._sign_up_anonymously.get_active()

    def is_advanced(self) -> bool:
        return self._sign_up_advanced_checkbutton.get_active()

    def get_server(self) -> str:
        return validate_domainpart(self._entry.get_text())

    def get_visible_buttons(self) -> list[str]:
        return ["back", "signup"]

    def get_default_button(self) -> str:
        return "signup"


@Gtk.Template(string=get_ui_string("wizard/recommendation_popover.ui"))
class WizardServerRecommendationPopover(Gtk.Popover, SignalManager):
    __gtype_name__ = "WizardServerRecommendationPopover"

    _visit_server_button: Gtk.Button = Gtk.Template.Child()
    _recommendation_link1: Gtk.Label = Gtk.Template.Child()
    _recommendation_link2: Gtk.Label = Gtk.Template.Child()

    def __init__(self) -> None:
        Gtk.Popover.__init__(self)
        SignalManager.__init__(self)

        self._website = ""

        self._connect(
            self._recommendation_link1, "activate-link", self._on_activate_link
        )
        self._connect(
            self._recommendation_link2, "activate-link", self._on_activate_link
        )
        self._connect(self._visit_server_button, "clicked", self._on_visit_server)

    def do_unroot(self) -> None:
        self._disconnect_all()
        Gtk.Popover.do_unroot(self)
        app.check_finalize(self)

    def set_website(self, website: str | None) -> None:
        self._website = website or ""
        self._visit_server_button.set_visible(bool(self._website))

    def _on_visit_server(self, _button: Gtk.Button) -> int:
        open_uri(f"https://{self._website}")
        return Gdk.EVENT_STOP

    @staticmethod
    def _on_activate_link(_label: Gtk.Label, uri: str) -> int:
        # We have to use this, because the default GTK handler
        # is not cross-platform compatible
        open_uri(uri)
        return Gdk.EVENT_STOP


@Gtk.Template(string=get_ui_string("wizard/advanced_page.ui"))
class WizardAdvancedPage(AssistantPage):
    __gtype_name__ = "WizardAdvancedPage"

    _custom_port_entry: Gtk.Entry = Gtk.Template.Child()
    _custom_host_entry: Gtk.Entry = Gtk.Template.Child()
    _proxies_dropdown: GajimDropDown[str] = Gtk.Template.Child()
    _manage_proxies_button: Gtk.Button = Gtk.Template.Child()
    _con_type_dropdown: GajimDropDown[str] = Gtk.Template.Child()

    def __init__(self) -> None:
        AssistantPage.__init__(self)
        self.title: str = _("Advanced settings")
        self.complete: bool = False

        self._connect(self._manage_proxies_button, "clicked", self._on_proxy_manager)
        self._connect(self._proxies_dropdown, "notify::selected", self._set_complete)
        self._connect(self._custom_host_entry, "changed", self._set_complete)
        self._connect(self._custom_port_entry, "changed", self._set_complete)

        self._con_type_dropdown.set_data(["START TLS", "DIRECT TLS", "PLAIN"])

        self.update_proxy_list()

    @staticmethod
    def _on_proxy_manager(_button: Gtk.Button) -> None:
        app.app.activate_action("manage-proxies", None)

    def update_proxy_list(self) -> None:
        proxies: dict[str, str] = {
            "": _("System"),
            "no-proxy": _("No Proxy"),
        }
        for proxy in app.settings.get_proxies():
            proxies[proxy] = proxy
        self._proxies_dropdown.set_data(proxies)

    def get_proxy(self) -> str:
        key = self._proxies_dropdown.get_selected_key()
        assert key is not None
        return key

    def get_custom_host(self) -> CustomHostT | None:
        host = self._custom_host_entry.get_text()
        port = self._custom_port_entry.get_text()
        if not host or not port:
            return None

        con_type = self._con_type_dropdown.get_selected_key()

        protocol = ConnectionProtocol.TCP
        if host.startswith(("ws://", "wss://")):
            protocol = ConnectionProtocol.WEBSOCKET

        return (f"{host}:{port}", protocol, ConnectionType(con_type))

    def _show_host_icon(self, show: bool) -> None:
        icon = "lucide-circle-alert-symbolic" if show else None
        self._custom_host_entry.set_icon_from_icon_name(
            Gtk.EntryIconPosition.SECONDARY, icon
        )

    def _show_port_icon(self, show: bool) -> None:
        icon = "lucide-circle-alert-symbolic" if show else None
        self._custom_port_entry.set_icon_from_icon_name(
            Gtk.EntryIconPosition.SECONDARY, icon
        )

    def _validate_host(self) -> bool:
        host = self._custom_host_entry.get_text()
        if host.startswith(("ws://", "wss://")):
            # We have no method for validating websocket URIs
            self._show_host_icon(False)
            return True

        try:
            validate_domainpart(host)
        except Exception:
            self._show_host_icon(True)
            self._custom_host_entry.set_icon_tooltip_text(
                Gtk.EntryIconPosition.SECONDARY, _("Invalid domain name")
            )
            return False

        self._show_host_icon(False)
        return True

    def _validate_port(self) -> bool:
        port = self._custom_port_entry.get_text()
        if not port:
            self._show_port_icon(False)
            return False

        try:
            port = int(port)
        except Exception:
            self._show_port_icon(True)
            self._custom_port_entry.set_icon_tooltip_text(
                Gtk.EntryIconPosition.SECONDARY, _("Must be a port number")
            )
            return False

        if port not in range(65535):
            self._show_port_icon(True)
            self._custom_port_entry.set_icon_tooltip_text(
                Gtk.EntryIconPosition.SECONDARY,
                _("Port must be a number between 0 and 65535"),
            )
            return False

        self._show_port_icon(False)
        return True

    def _is_custom_host_set(self) -> bool:
        host = bool(self._custom_host_entry.get_text())
        port = bool(self._custom_port_entry.get_text())
        return host or port

    def _is_proxy_set(self) -> bool:
        return self._proxies_dropdown.get_selected_key() != ""

    def _set_complete(self, *args: Any) -> None:
        self.complete = False
        if self._is_proxy_set():
            self.complete = True

        if self._is_custom_host_set():
            port_valid = self._validate_port()
            host_valid = self._validate_host()
            self.complete = port_valid and host_valid

        self.update_page_complete()

    def get_visible_buttons(self) -> list[str]:
        window = get_app_window("AccountWizard")
        assert window is not None
        return ["back", window.get_current_method()]

    def get_default_button(self) -> str:
        window = get_app_window("AccountWizard")
        assert window is not None
        return window.get_current_method()


@Gtk.Template(string=get_ui_string("wizard/security_page.ui"))
class WizardSecurityPage(AssistantPage):
    __gtype_name__ = "WizardSecurityPage"

    _error_list: Gtk.ListBox = Gtk.Template.Child()
    _view_cert_button: Gtk.Button = Gtk.Template.Child()
    _trust_cert_checkbutton: Gtk.CheckButton = Gtk.Template.Child()

    def __init__(self) -> None:
        AssistantPage.__init__(self)
        self.title: str = _("Security Warning")
        self._cert: Gio.TlsCertificate | None = None

        self._connect(self._view_cert_button, "clicked", self._on_view_cert)

    @property
    def cert(self) -> Gio.TlsCertificate | None:
        return self._cert

    def set_warning(
        self,
        cert: Gio.TlsCertificate,
        errors: list[Gio.TlsCertificateFlags],
    ) -> None:
        self._cert = cert
        clear_listbox(self._error_list)

        unknown_error = _('Unknown TLS error "%s"')
        for error in errors:
            error_text = GIO_TLS_ERRORS.get(error, unknown_error % error)
            box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
            image = Gtk.Image.new_from_icon_name("lucide-circle-alert-symbolic")
            image.add_css_class("warning")
            label = Gtk.Label(
                label=error_text,
                wrap=True,
                xalign=0,
                selectable=True,
            )
            box.append(image)
            box.append(label)
            self._error_list.append(box)

        self._trust_cert_checkbutton.set_visible(
            Gio.TlsCertificateFlags.UNKNOWN_CA in errors
        )

    def _on_view_cert(self, _button: Gtk.Button) -> None:
        open_window(
            "CertificateDialog",
            account=None,
            transient_for=self.get_root(),
            cert=self._cert,
        )

    @property
    def trust_certificate(self) -> bool:
        return self._trust_cert_checkbutton.get_active()

    def get_visible_buttons(self) -> list[str]:
        window = get_app_window("AccountWizard")
        assert window is not None
        return ["back", window.get_current_method()]

    def get_default_button(self) -> str:
        return "back"


class WizardFormPage(AssistantPage):
    def __init__(self) -> None:
        AssistantPage.__init__(self)
        self.set_valign(Gtk.Align.FILL)
        self.complete: bool = False
        self.title: str = _("Create Account")
        self._current_form: DataFormWidget | None = None

        heading = Gtk.Label(
            label=_("Create Account"),
            wrap=True,
            max_width_chars=30,
            halign=Gtk.Align.CENTER,
            justify=Gtk.Justification.CENTER,
        )
        heading.add_css_class("title-1")
        self.append(heading)

    @property
    def has_form(self) -> bool:
        return self._current_form is not None

    def _on_is_valid(self, _widget: DataFormWidget, is_valid: bool) -> None:
        try:
            self.get_credentials()
        except Exception:
            is_valid = False

        self.complete = is_valid
        self.update_page_complete()

    def add_form(self, form: Any) -> None:
        self.remove_form()

        options = {"hide-fallback-fields": True, "entry-activates-default": True}
        self._current_form = DataFormWidget(form, options)
        self._connect(self._current_form, "is-valid", self._on_is_valid)
        self._current_form.validate()
        self.append(self._current_form)

    def get_credentials(self) -> tuple[str, str]:
        assert self._current_form is not None
        form = cast(SimpleDataForm, self._current_form.get_form())
        username = cast(str, form["username"].value)
        password = cast(str, form["password"].value)
        return username, password

    def get_submit_form(self) -> Any:
        assert self._current_form is not None
        return self._current_form.get_submit_form()

    def remove_form(self) -> None:
        if self._current_form is None:
            return

        self.remove(self._current_form)
        self._current_form = None

    def focus(self) -> None:
        assert self._current_form is not None
        self._current_form.focus_first_entry()

    def get_visible_buttons(self) -> list[str]:
        return ["back", "signup"]

    def get_default_button(self) -> str:
        return "signup"


@Gtk.Template(string=get_ui_string("wizard/redirect_page.ui"))
class WizardRedirectPage(AssistantPage):
    __gtype_name__ = "WizardRedirectPage"

    _instructions: Gtk.Label = Gtk.Template.Child()
    _link_button: Gtk.Button = Gtk.Template.Child()

    def __init__(self) -> None:
        AssistantPage.__init__(self)
        self.title: str = _("Redirect")
        self._link: str | None = None

        self._connect(self._link_button, "clicked", self._on_link_button)

    def set_redirect(self, link: str, instructions: str | None) -> None:
        if instructions is None:
            instructions = _("Register on the Website")
        self._instructions.set_text(instructions)
        self._link = link

    def _on_link_button(self, _button: Gtk.Button) -> None:
        assert self._link is not None
        open_uri(self._link)

    def get_visible_buttons(self) -> list[str]:
        return ["back"]


@Gtk.Template(string=get_ui_string("wizard/success_page.ui"))
class WizardSuccessPage(AssistantSuccessPage):
    __gtype_name__ = "WizardSuccessPage"

    _account_name_entry: Gtk.Entry = Gtk.Template.Child()
    _account_color_button: Gtk.ColorDialogButton = Gtk.Template.Child()
    _badge_preview: Gtk.Label = Gtk.Template.Child()

    def __init__(self) -> None:
        AssistantSuccessPage.__init__(self)
        self.set_title(_("Account Added"))
        self.set_heading(_("Account has been added successfully"))

        self._account: str | None = None
        self._our_jid: str | None = None
        self._label: str | None = None
        self._color: str | None = None

        self._provider = self._add_css_provider()

        self._connect(self._account_name_entry, "changed", self._on_name_changed)

        color_dialog = Gtk.ColorDialog()
        self._account_color_button.set_dialog(color_dialog)
        self._connect(self._account_color_button, "notify::rgba", self._on_color_set)

    def set_account(self, account: str) -> None:
        self._account = account
        self._our_jid = app.get_jid_from_account(account)
        self._badge_preview.set_text(self._our_jid)
        rgba = Gdk.RGBA()
        rgba.parse(get_color_for_account(self._our_jid))
        self._account_color_button.set_rgba(rgba)
        self._color = rgba.to_string()
        self._set_badge_color(self._color)
        self._save_config()

    @property
    def account(self) -> str | None:
        return self._account

    def _add_css_provider(self) -> Gtk.CssProvider:
        context = self._badge_preview.get_style_context()
        provider = Gtk.CssProvider()
        context.add_provider(provider, Gtk.STYLE_PROVIDER_PRIORITY_USER)
        return provider

    def _on_name_changed(self, entry: Gtk.Entry):
        assert self._our_jid is not None
        self._label = entry.get_text()
        self._badge_preview.set_text(self._label or self._our_jid)
        self._save_config()

    def _on_color_set(self, color_button: Gtk.ColorDialogButton, *args: Any):
        rgba = color_button.get_rgba()
        self._color = rgba.to_string()
        self._set_badge_color(self._color)
        self._save_config()

    def _set_badge_color(self, color: str) -> None:
        css = ".badge { background-color: %s; font-size: 100%%; }" % color
        self._provider.load_from_bytes(GLib.Bytes.new(css.encode("utf-8")))

    def _save_config(self) -> None:
        assert self._account is not None
        app.settings.set_account_setting(self._account, "account_color", self._color)
        if self._label:
            app.settings.set_account_setting(
                self._account, "account_label", self._label
            )
        app.css_config.refresh()

    def get_visible_buttons(self) -> list[str]:
        return ["connect"]


class WizardErrorPage(AssistantErrorPage):
    def __init__(self) -> None:
        AssistantErrorPage.__init__(self)
        self.set_heading(_("An error occurred during account creation"))

    def get_visible_buttons(self) -> list[str]:
        return ["back"]



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast

import logging

from gi.repository import Gio
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango

from gajim.common import app
from gajim.common.client import Client
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.storage.archive.models import Occupant

from gajim.gtk.builder import get_builder
from gajim.gtk.util.classes import SignalManager

log = logging.getLogger("gajim.gtk.groupchat_blocks")


class GroupchatBlocks(Gtk.Box, SignalManager):
    def __init__(self, client: Client, contact: GroupchatContact) -> None:
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)
        SignalManager.__init__(self)

        self._client = client
        self._account = contact.account
        self._contact = contact
        self._model = Gio.ListStore(item_type=BlockListItem)

        self._ui = get_builder("groupchat_blocks.ui")
        self.append(self._ui.main)

        for occupant in app.storage.archive.get_blocked_occupants(
            contact.account, contact.jid
        ):
            self._model.append(BlockListItem(occupant))

        columns = [
            (self._ui.nickname_col, BlockViewItemLabel, "nickname"),
            (self._ui.id_col, BlockViewItemLabel, "id"),
        ]

        for col, widget, attr in columns:
            factory = col.get_factory()
            assert factory is not None
            self._connect(factory, "setup", self._on_factory_setup, widget)
            self._connect(factory, "bind", self._on_factory_bind, attr)
            self._connect(factory, "unbind", self._on_factory_unbind)

        expression = Gtk.PropertyExpression.new(
            this_type=BlockListItem,
            expression=None,
            property_name="nickname",
        )
        sorter = Gtk.StringSorter.new(expression=expression)
        sort_model = Gtk.SortListModel(model=self._model, sorter=sorter)

        expression = Gtk.PropertyExpression.new(
            this_type=BlockListItem,
            expression=None,
            property_name="nickname",
        )

        self._string_filter = Gtk.StringFilter.new(expression)

        filter_model = Gtk.FilterListModel(model=sort_model, filter=self._string_filter)

        self._selection_model = Gtk.MultiSelection(model=filter_model)
        self._ui.column_view.set_model(self._selection_model)

        self._connect(self._ui.search_entry, "search-changed", self._on_search_changed)
        self._connect(self._ui.remove_button, "clicked", self._on_remove_button_clicked)

    def _cleanup(self, *args: Any) -> None:
        self._client.disconnect_all_from_obj(self)
        del self._selection_model
        del self._model
        del self._string_filter
        del self._client
        del self._contact

    @staticmethod
    def _on_factory_setup(
        _factory: Gtk.SignalListItemFactory, list_item: Gtk.ListItem, item_class: Any
    ) -> None:
        list_item.set_child(item_class())

    @staticmethod
    def _on_factory_bind(
        _factory: Gtk.SignalListItemFactory, list_item: Gtk.ListItem, attr: str
    ) -> None:
        block_item = cast(BlockListItem, list_item.get_item())
        block_view_item = cast(BlockViewItemLabel, list_item.get_child())
        block_view_item.bind(block_item, attr)

    @staticmethod
    def _on_factory_unbind(
        _factory: Gtk.SignalListItemFactory, list_item: Gtk.ListItem
    ) -> None:
        block_view_item = cast(BlockViewItemLabel, list_item.get_child())
        block_view_item.unbind()

    def _remove(self, item: BlockListItem) -> None:
        for index, item_ in enumerate(self._model):
            if item_.id == item.id:
                self._model.remove(index)
                return

    def _get_selected_items(self) -> list[BlockListItem]:
        bitset = self._selection_model.get_selection()
        valid, iter_, value = Gtk.BitsetIter.init_first(bitset)

        items: list[BlockListItem] = []

        if not valid:
            return items

        items.append(cast(BlockListItem, self._selection_model.get_item(value)))

        while res := iter_.next():
            valid, value = res
            if not valid:
                break

            items.append(cast(BlockListItem, self._selection_model.get_item(value)))

        return items

    def _on_search_changed(self, search_entry: Gtk.SearchEntry) -> None:
        self._string_filter.set_search(search_entry.get_text())

    def _on_remove_button_clicked(self, _button: Gtk.Button) -> None:
        items = self._get_selected_items()
        if not items:
            return

        occupant_ids: list[str] = []
        for item in items:
            occupant_ids.append(item.id)
            self._remove(item)

        self._client.get_module("MucBlocking").set_block_occupants(
            self._contact.jid, occupant_ids, False
        )


class BlockListItem(GObject.Object):
    __gtype_name__ = "BlockListItem"

    id = GObject.Property(type=str)
    nickname = GObject.Property(type=str)

    def __init__(self, occupant: Occupant) -> None:
        super().__init__(
            id=occupant.id,
            nickname=occupant.nickname or "Unknown",
        )

    def __repr__(self) -> str:
        return f"BlockListItem: {self.id} {self.nickname}"


class BlockViewItemLabel(Gtk.Label):
    __gtype_name__ = "BlockViewItemLabel"

    def __init__(self) -> None:
        Gtk.Label.__init__(
            self,
            ellipsize=Pango.EllipsizeMode.END,
            width_chars=20,
            xalign=0,
        )

        self.__bindings: list[GObject.Binding] = []

    def bind(self, obj: BlockListItem, attr: str) -> None:
        bind_spec = [
            (attr, self, "label"),
        ]

        for source_prop, widget, target_prop in bind_spec:
            bind = obj.bind_property(
                source_prop, widget, target_prop, GObject.BindingFlags.SYNC_CREATE
            )
            self.__bindings.append(bind)

    def unbind(self) -> None:
        for bind in self.__bindings:
            bind.unbind()
        self.__bindings.clear()

    def do_unroot(self) -> None:
        Gtk.Label.do_unroot(self)
        app.check_finalize(self)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast
from typing import Literal

import calendar
import datetime as dt
import itertools
import logging
import operator
import pprint
import shutil
from collections.abc import Iterable
from collections.abc import Iterator
from collections.abc import Sequence
from datetime import datetime
from datetime import timedelta
from pathlib import Path

import sqlalchemy as sa
from nbxmpp import JID
from sqlalchemy import delete
from sqlalchemy import func
from sqlalchemy import Row
from sqlalchemy import select
from sqlalchemy import update
from sqlalchemy.dialects.sqlite import insert
from sqlalchemy.engine import CursorResult
from sqlalchemy.engine import Engine
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import aliased
from sqlalchemy.orm import contains_eager
from sqlalchemy.orm import joinedload
from sqlalchemy.orm import selectinload
from sqlalchemy.orm import Session

from gajim.common import app
from gajim.common import configpaths
from gajim.common.const import MAX_MESSAGE_CORRECTION_DELAY
from gajim.common.events import DBMigration
from gajim.common.storage.archive import migration
from gajim.common.storage.archive.const import ChatDirection
from gajim.common.storage.archive.const import MessageState
from gajim.common.storage.archive.const import MessageType
from gajim.common.storage.archive.models import Account
from gajim.common.storage.archive.models import Base
from gajim.common.storage.archive.models import DisplayedMarker
from gajim.common.storage.archive.models import Encryption
from gajim.common.storage.archive.models import MAMArchiveState
from gajim.common.storage.archive.models import Message
from gajim.common.storage.archive.models import MessageError
from gajim.common.storage.archive.models import Moderation
from gajim.common.storage.archive.models import Occupant
from gajim.common.storage.archive.models import OpenGraph
from gajim.common.storage.archive.models import Reaction
from gajim.common.storage.archive.models import Receipt
from gajim.common.storage.archive.models import Remote
from gajim.common.storage.archive.models import Retraction
from gajim.common.storage.archive.models import SecurityLabel
from gajim.common.storage.archive.models import Thread
from gajim.common.storage.base import AlchemyStorage
from gajim.common.storage.base import timeit
from gajim.common.storage.base import VALUE_MISSING
from gajim.common.storage.base import with_session
from gajim.common.storage.base import with_session_yield_from
from gajim.common.util.datetime import FIRST_UTC_DATETIME
from gajim.common.util.datetime import utc_now
from gajim.common.util.text import get_random_string

CURRENT_USER_VERSION = 17


log = logging.getLogger("gajim.c.storage.archive")


class MessageArchiveStorage(AlchemyStorage):
    def __init__(self, in_memory: bool = False, path: Path | None = None) -> None:
        if path is None:
            path = configpaths.get("LOG_DB")

        AlchemyStorage.__init__(
            self,
            log,
            None if in_memory else path,
            pragma={
                "journal_mode": "wal",
                "secure_delete": "on",
            },
        )

        self._account_pks: dict[str, int] = {}
        self._jid_pks: dict[JID, int] = {}
        self._occupant_cache: dict[tuple[str, JID, JID], tuple[Occupant, datetime]] = {}

    def init(self) -> None:
        super().init()
        with self._session as s:
            self._load_jids(s)

    def _log_row(self, row: Any) -> None:
        if self._log.getEffectiveLevel() != logging.DEBUG:
            return
        self._log.debug("Object before query\n%s", pprint.pformat(row))

    def _create_table(self, session: Session, engine: Engine) -> None:
        Base.metadata.create_all(engine)
        session.execute(sa.text(f"PRAGMA user_version={CURRENT_USER_VERSION}"))

    def _make_backup(self) -> None:
        db_path = configpaths.get("LOG_DB")
        random_string = get_random_string(10)
        db_backup_path = db_path.parent / f"{db_path.name}.{random_string}.bak"
        shutil.copy(db_path, db_backup_path)

    def _migrate(self) -> None:
        user_version = self._get_user_version()
        if user_version < CURRENT_USER_VERSION:
            app.ged.raise_event(DBMigration())
            self._make_backup()
            migration.run(self, user_version)

    @timeit
    def _load_jids(self, session: Session) -> None:
        jids = session.scalars(select(Remote))
        self._jid_pks = {j.jid: j.pk for j in jids}

    def _get_active_account_pks(self, session: Session) -> list[int]:
        accounts = app.settings.get_active_accounts()
        return [self._get_account_pk(session, account) for account in accounts]

    def _get_account_pk(self, session: Session, account: str) -> int:
        pk = self._account_pks.get(account)
        if pk is not None:
            return pk

        jid_str = app.get_jid_from_account(account)
        jid = JID.from_string(jid_str)

        pk = session.scalar(select(Account.pk).where(Account.jid == jid))
        if pk is None:
            acc = Account(jid=jid)
            session.add(acc)
            session.flush()
            pk = acc.pk

        self._account_pks[account] = pk
        return pk

    def _get_jid_pk(self, session: Session, jid: JID) -> int:
        pk = self._jid_pks.get(jid)
        if pk is not None:
            return pk

        jid_row = Remote(jid=jid)
        session.add(jid_row)
        session.flush()

        pk = jid_row.pk
        self._jid_pks[jid] = pk
        return pk

    def _set_foreign_keys(self, session: Session, row: Any) -> None:
        fk_account_pk = None
        account = getattr(row, "account_", None)
        if account is not None:
            fk_account_pk = self._get_account_pk(session, account)
            row.account_ = None
            row.fk_account_pk = fk_account_pk

        fk_remote_pk = None
        remote_jid = getattr(row, "remote_jid_", None)
        if remote_jid is not None:
            row.remote_jid_ = None
            fk_remote_pk = self._get_jid_pk(session, remote_jid)
            row.fk_remote_pk = fk_remote_pk

        if hasattr(row, "real_remote_jid_"):
            real_remote_jid = row.real_remote_jid_
            if real_remote_jid is not VALUE_MISSING:
                row.real_remote_jid_ = None
                if real_remote_jid is None:
                    row.fk_real_remote_pk = None
                else:
                    row.fk_real_remote_pk = self._get_jid_pk(session, real_remote_jid)

        security_label = getattr(row, "security_label_", None)
        if security_label is not None:
            pk = self._upsert_row(session, row.security_label_)
            row.security_label_ = None
            row.fk_security_label_pk = pk

        encryption = getattr(row, "encryption_", None)
        if encryption is not None:
            pk = self._insert_row(session, encryption, return_pk_on_conflict=True)
            row.encryption_ = None
            row.fk_encryption_pk = pk

        thread_id = getattr(row, "thread_id_", None)
        if thread_id is not None:
            assert fk_account_pk is not None
            assert fk_remote_pk is not None
            thread = Thread(
                fk_account_pk=fk_account_pk,
                fk_remote_pk=fk_remote_pk,
                id=thread_id,
            )
            pk = self._insert_row(session, thread, return_pk_on_conflict=True)
            row.thread_ = None
            row.fk_thread_pk = pk

        occupant = getattr(row, "occupant_", None)
        if occupant is not None:
            pk = self._upsert_row(session, occupant)
            row.occupant_ = None
            row.fk_occupant_pk = pk

    @with_session
    @timeit
    def insert_object(
        self, session: Session, obj: Any, ignore_on_conflict: bool = True
    ) -> int:
        obj.validate()
        self._set_foreign_keys(session, obj)
        self._log_row(obj)
        session.add(obj)

        try:
            session.commit()
        except Exception:
            if not ignore_on_conflict:
                raise
            return -1

        return obj.pk

    @with_session
    @timeit
    def insert_row(
        self,
        session: Session,
        row: Any,
        *,
        return_pk_on_conflict: bool = False,
        ignore_on_conflict: bool = False,
    ) -> int:
        return self._insert_row(
            session,
            row,
            return_pk_on_conflict=return_pk_on_conflict,
            ignore_on_conflict=ignore_on_conflict,
        )

    def _insert_row(
        self,
        session: Session,
        row: Any,
        *,
        return_pk_on_conflict: bool = False,
        ignore_on_conflict: bool = False,
    ) -> int:
        row.validate()
        self._set_foreign_keys(session, row)
        self._log_row(row)
        table = row.__class__
        stmt = insert(table).values(**row.get_insert_values()).returning(table.pk)

        try:
            pk = session.scalar(stmt)
        except IntegrityError:
            if ignore_on_conflict:
                return -1

            if return_pk_on_conflict:
                stmt = row.get_select_stmt()
                return session.scalar(stmt).pk

            raise

        assert pk is not None
        return pk

    @with_session
    @timeit
    def upsert_row(self, session: Session, row: Any) -> int:
        return self._upsert_row(session, row)

    def _upsert_row(
        self,
        session: Session,
        row: Any,
    ) -> int:
        row.validate()
        self._set_foreign_keys(session, row)
        self._log_row(row)
        table = row.__class__

        stmt = row.get_select_stmt()
        existing = session.scalar(stmt)

        if existing is None:
            stmt = insert(table).values(**row.get_insert_values()).returning(table.pk)
            pk = session.scalar(stmt)
            assert pk is not None
            return pk

        if not row.needs_update(existing):
            return existing.pk

        stmt = (
            update(table)
            .where(table.pk == existing.pk)
            .values(**row.get_upsert_values())
        )

        session.execute(stmt)
        return existing.pk

    @with_session
    @timeit
    def upsert_row2(self, session: Session, row: Any) -> int | None:
        return self._upsert_row2(session, row)

    def _upsert_row2(
        self,
        session: Session,
        row: Any,
    ) -> int | None:
        row.validate()
        self._set_foreign_keys(session, row)
        self._log_row(row)
        table = row.__class__

        stmt = insert(table).values(**row.get_insert_values())
        stmt = stmt.on_conflict_do_update(
            set_=row.get_upsert_values(),
            where=sa.text("excluded.timestamp > timestamp"),
        )
        stmt = stmt.returning(table.pk)
        pk = session.scalar(stmt)
        return pk

    @with_session
    @timeit
    def get_message_with_pk(
        self, session: Session, pk: int, options: Any = None
    ) -> Message | None:
        return self._get_message_with_pk(session, pk, options)

    def _get_message_with_pk(
        self, session: Session, pk: int, options: Any = None
    ) -> Message | None:
        stmt = select(Message).where(Message.pk == pk)
        if options is not None:
            stmt = stmt.options(*options)
        return session.scalar(stmt)

    @with_session
    @timeit
    def get_message_with_id(
        self, session: Session, account: str, jid: JID, message_id: str
    ) -> Message | None:
        fk_account_pk = self._get_account_pk(session, account)
        fk_remote_pk = self._get_jid_pk(session, jid)

        stmt = select(Message).where(
            Message.id == message_id,
            Message.fk_remote_pk == fk_remote_pk,
            Message.fk_account_pk == fk_account_pk,
        )

        result = session.scalars(stmt).all()
        if len(result) == 1:
            return result[0]

        self._log.warning(
            "Found %s messages with message id %s", len(result), message_id
        )
        return None

    @with_session
    @timeit
    def get_message_with_stanza_id(
        self, session: Session, account: str, jid: JID, stanza_id: str
    ) -> Message | None:
        fk_account_pk = self._get_account_pk(session, account)
        fk_remote_pk = self._get_jid_pk(session, jid)

        stmt = select(Message).where(
            Message.stanza_id == stanza_id,
            Message.fk_remote_pk == fk_remote_pk,
            Message.fk_account_pk == fk_account_pk,
        )

        result = session.scalars(stmt).all()
        if len(result) == 1:
            return result[0]

        self._log.warning("Found %s messages with stanza id %s", len(result), stanza_id)
        return None

    @with_session
    @timeit
    def delete_message(self, session: Session, pk: int) -> None:
        message = self._get_message_with_pk(
            session, pk, options=[selectinload(Message.markers)]
        )
        if message is None:
            self._log.warning("Deletion failed, no message found with pk %s", pk)
            return

        self._delete_message(session, message)

    def _delete_message(self, session: Session, message: Message) -> None:
        if message.corrections:
            for correction in message.corrections:
                session.delete(correction)

        if message.error is not None:
            session.delete(message.error)

        if message.moderation is not None:
            session.delete(message.moderation)

        if message.retraction is not None:
            session.delete(message.retraction)

        if message.receipt is not None:
            session.delete(message.receipt)

        for marker in message.markers:
            session.delete(marker)

        for reaction in message.reactions:
            session.delete(reaction)

        session.delete(message)

    @with_session
    @timeit
    def delete_reaction(
        self,
        session: Session,
        account: str,
        jid: JID,
        occupant_id: str | None,
        reaction_id: str,
        direction: ChatDirection,
    ) -> None:
        fk_account_pk = self._get_account_pk(session, account)
        fk_remote_pk = self._get_jid_pk(session, jid)

        fk_occupant_pk = None
        if occupant_id is not None:
            stmt = select(Occupant.pk).where(
                Occupant.id == occupant_id,
                Occupant.fk_remote_pk == fk_remote_pk,
                Occupant.fk_account_pk == fk_account_pk,
            )
            fk_occupant_pk = session.scalar(stmt)
            if fk_occupant_pk is None:
                self._log.warning(
                    "Unable to delete reaction, unknown occupant-id: %s", occupant_id
                )
                return

        stmt = delete(Reaction).where(
            Reaction.id == reaction_id,
            Reaction.fk_remote_pk == fk_remote_pk,
        )

        if fk_occupant_pk is None:
            stmt = stmt.where(Reaction.fk_occupant_pk.is_(None))
        else:
            stmt = stmt.where(Reaction.fk_occupant_pk == fk_occupant_pk)

        stmt = stmt.where(
            Reaction.fk_account_pk == fk_account_pk,
            Reaction.direction == direction,
        )

        session.execute(stmt)

    @with_session
    @timeit
    def check_if_message_id_exists(
        self, session: Session, account: str, jid: JID, message_id: str
    ) -> bool:
        fk_account_pk = self._get_account_pk(session, account)
        fk_remote_pk = self._get_jid_pk(session, jid)

        exists_criteria = (
            select(Message.id)
            .where(
                Message.id == message_id,
                Message.fk_remote_pk == fk_remote_pk,
                Message.fk_account_pk == fk_account_pk,
            )
            .exists()
        )

        res = session.scalar(select(1).where(exists_criteria))
        return bool(res)

    @with_session
    @timeit
    def check_if_stanza_id_exists(
        self, session: Session, account: str, jid: JID, stanza_id: str
    ) -> bool:
        fk_account_pk = self._get_account_pk(session, account)
        fk_remote_pk = self._get_jid_pk(session, jid)

        exists_criteria = (
            select(Message.id)
            .where(
                Message.stanza_id == stanza_id,
                Message.fk_remote_pk == fk_remote_pk,
                Message.fk_account_pk == fk_account_pk,
            )
            .exists()
        )

        res = session.scalar(select(1).where(exists_criteria))
        return bool(res)

    @with_session
    @timeit
    def get_conversation_jids(
        self, session: Session, account: str
    ) -> Sequence[Row[tuple[JID, int]]]:
        fk_account_pk = self._get_account_pk(session, account)

        subq = (
            select(Message.fk_remote_pk, Message.type)
            .distinct()
            .where(Message.fk_account_pk == fk_account_pk)
        ).subquery()

        stmt = select(Remote.jid, subq.c.type).join(subq)
        self._explain(session, stmt)
        return session.execute(stmt).fetchall()

    @with_session
    @timeit
    def get_conversation_before_after(
        self,
        session: Session,
        account: str,
        jid: JID,
        timestamp: datetime,
        n_lines: int,
        *,
        direction: Literal["before", "after"],
        order: Literal["asc", "desc"] = "asc",
        include_timestamp: bool = False,
    ) -> tuple[Iterable[Message], bool]:
        """
        Load n messages from jid before or after timestamp

        :param account:
            The account
        :param jid:
            The jid for which we request the conversation
        :param timestamp:
            The point in time from where to search
        :param nlines:
            The maximal count of Message returned
        :param direction:
            The search direction
        :param order:
            How the result is ordered
        :param include_timestamp:
            If messages with the same timestamp are returned
        """

        fk_account_pk = self._get_account_pk(session, account)
        fk_remote_pk = self._get_jid_pk(session, jid)

        stmt = select(Message).where(
            Message.fk_remote_pk == fk_remote_pk,
            Message.fk_account_pk == fk_account_pk,
            Message.correction_id.is_(None),
        )

        if direction == "before":
            if include_timestamp:
                where = stmt.where(Message.timestamp <= timestamp)
            else:
                where = stmt.where(Message.timestamp < timestamp)
            stmt = where.order_by(sa.desc(Message.timestamp), sa.desc(Message.pk))

        else:
            if include_timestamp:
                where = stmt.where(Message.timestamp >= timestamp)
            else:
                where = stmt.where(Message.timestamp > timestamp)
            stmt = where.order_by(Message.timestamp, Message.pk)

        stmt = stmt.limit(n_lines)

        self._explain(session, stmt)
        result = session.scalars(stmt).all()
        complete = len(result) < n_lines

        match order, direction:
            case ("desc", "before") | ("asc", "after"):
                return result, complete
            case ("desc", "after") | ("asc", "before"):
                return reversed(result), complete

    @with_session
    @timeit
    def get_conversation_around_timestamp(
        self, session: Session, account: str, jid: JID, timestamp: datetime
    ) -> tuple[Iterable[Message], bool, bool]:
        """
        Loads messages around a primary key

        :param account:
            The account
        :param jid:
            The jid for which we request the conversation
        :param timestamp:
            The timestamp in the conversation
        """

        messages_before, before_complete = self.get_conversation_before_after(
            account,
            jid,
            timestamp,
            50,
            direction="before",
            order="asc",
            include_timestamp=True,
        )
        messages_after, after_complete = self.get_conversation_before_after(
            account, jid, timestamp, 50, direction="after"
        )

        return (
            itertools.chain(messages_before, messages_after),
            before_complete,
            after_complete,
        )

    @with_session
    @timeit
    def get_last_conversation_row(
        self, session: Session, account: str, jid: JID
    ) -> Message | None:
        """
        Load the last line of a conversation with jid for account.
        Loads messages, but no status messages or error messages.

        :param account:         The account
        :param jid:             The jid for which we request the conversation

        returns a namedtuple or None
        """

        fk_account_pk = self._get_account_pk(session, account)
        fk_remote_pk = self._get_jid_pk(session, jid)

        stmt = (
            select(Message)
            .where(
                Message.fk_remote_pk == fk_remote_pk,
                Message.fk_account_pk == fk_account_pk,
                Message.correction_id.is_(None),
                sa.or_(Occupant.blocked == sa.false(), Occupant.blocked.is_(None)),
            )
            .outerjoin(Occupant)
            .options(contains_eager(Message.occupant))
            .order_by(sa.desc(Message.timestamp), sa.desc(Message.pk))
            .limit(1)
        )

        self._explain(session, stmt)
        return session.scalar(stmt)

    @with_session
    @timeit
    def get_last_correctable_message(
        self, session: Session, account: str, jid: JID, message_id: str
    ) -> Message | None:
        """
        Load the last correctable message of a conversation by message_id.
        """

        fk_account_pk = self._get_account_pk(session, account)
        fk_remote_pk = self._get_jid_pk(session, jid)

        min_time = datetime.now(dt.UTC) - timedelta(
            seconds=MAX_MESSAGE_CORRECTION_DELAY
        )

        stmt = (
            select(Message)
            .where(
                Message.id == message_id,
                Message.fk_remote_pk == fk_remote_pk,
                Message.fk_account_pk == fk_account_pk,
                Message.direction == ChatDirection.OUTGOING,
                Message.timestamp > min_time,
                Message.state == MessageState.ACKNOWLEDGED,
            )
            .order_by(sa.desc(Message.timestamp), sa.desc(Message.pk))
            .limit(1)
        )

        self._explain(session, stmt)
        return session.scalar(stmt)

    @with_session
    @timeit
    def get_corrected_message(
        self, session: Session, correction: Message
    ) -> Message | None:
        stmt = select(Message).where(
            Message.id == correction.correction_id,
            Message.fk_remote_pk == correction.fk_remote_pk,
            Message.fk_account_pk == correction.fk_account_pk,
            Message.fk_occupant_pk.is_(correction.fk_occupant_pk),
            Message.direction == correction.direction,
        )

        if correction.type == 2 and correction.fk_occupant_pk is None:
            stmt = stmt.where(Message.resource == correction.resource)

        stmt = stmt.order_by(sa.desc(Message.timestamp)).limit(1)

        self._explain(session, stmt)
        return session.scalar(stmt)

    def get_referenced_message(
        self, account: str, jid: JID, message_type: MessageType | int, reply_id: str
    ) -> Message | None:
        if message_type == MessageType.GROUPCHAT:
            return app.storage.archive.get_message_with_stanza_id(
                account, jid, reply_id
            )

        return app.storage.archive.get_message_with_id(account, jid, reply_id)

    @with_session
    @timeit
    def get_display_marker_with_pk(
        self, session: Session, pk: int
    ) -> DisplayedMarker | None:
        stmt = (
            select(DisplayedMarker)
            .where(DisplayedMarker.pk == pk)
            .options(joinedload(DisplayedMarker.remote))
        )
        return session.scalar(stmt)

    @with_session
    @timeit
    def get_last_display_marker(
        self, session: Session, account: str, jid: JID
    ) -> DisplayedMarker | None:
        fk_account_pk = self._get_account_pk(session, account)
        fk_remote_pk = self._get_jid_pk(session, jid)

        stmt = (
            select(DisplayedMarker)
            .where(
                DisplayedMarker.fk_remote_pk == fk_remote_pk,
                DisplayedMarker.fk_account_pk == fk_account_pk,
                DisplayedMarker.timestamp > utc_now() - timedelta(days=180),
            )
            .order_by(sa.desc(DisplayedMarker.timestamp))
            .limit(1)
            .options(joinedload(DisplayedMarker.remote))
        )
        self._explain(session, stmt)
        return session.scalar(stmt)

    @with_session
    @timeit
    def get_last_display_markers(
        self, session: Session, account: str, jid: JID
    ) -> list[DisplayedMarker]:
        fk_account_pk = self._get_account_pk(session, account)
        fk_remote_pk = self._get_jid_pk(session, jid)

        cte = (
            select(
                DisplayedMarker,
                func.row_number()
                .over(
                    partition_by=DisplayedMarker.fk_occupant_pk,
                    order_by=sa.desc(DisplayedMarker.timestamp),
                )
                .label("rn"),
            )
            .where(
                DisplayedMarker.fk_remote_pk == fk_remote_pk,
                DisplayedMarker.fk_account_pk == fk_account_pk,
                DisplayedMarker.timestamp > utc_now() - timedelta(days=90),
            )
            .cte()
        )

        dm = aliased(DisplayedMarker, cte)
        stmt = select(dm).where(cte.c.rn == 1).options(joinedload(dm.remote))

        self._explain(session, stmt)
        return list(session.scalars(stmt).all())

    @with_session
    @timeit
    def get_message_stanza_ids_from_occupant(
        self,
        session: Session,
        account: str,
        jid: JID,
        occupant_id: str,
    ) -> Sequence[str] | None:
        """Get stanza IDs for all messages which were sent by a participant
        with occupant_id.
        """
        fk_account_pk = self._get_account_pk(session, account)
        fk_remote_pk = self._get_jid_pk(session, jid)
        stmt = select(Occupant.pk).where(
            Occupant.id == occupant_id,
            Occupant.fk_remote_pk == fk_remote_pk,
            Occupant.fk_account_pk == fk_account_pk,
        )
        fk_occupant_pk = session.scalar(stmt)
        if fk_occupant_pk is None:
            self._log.warning(
                "Unable to get messages, unknown occupant-id: %s", occupant_id
            )
            return None

        stmt = select(Message.stanza_id).where(
            Message.fk_remote_pk == fk_remote_pk,
            Message.fk_account_pk == fk_account_pk,
            Message.fk_occupant_pk == fk_occupant_pk,
            Message.stanza_id.isnot(None),
            ~Message.moderation.has(),
        )
        result = cast(Sequence[str] | None, session.scalars(stmt).all())
        return result

    @with_session_yield_from
    @timeit
    def search_archive(
        self,
        session: Session,
        account: str | None,
        jid: JID | None,
        query: str,
        from_users: list[str] | None = None,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> Iterator[Message]:
        """
        Search the conversation log for messages containing the `query` string.

        The search can either span the complete log for the given
        `account` and `jid` or be restricted to a single day by
        specifying `date`.

        :param account: The account

        :param jid: The jid for which we request the conversation

        :param query: A search string

        :param from_users: A list of usernames or None

        :param before: A datetime.datetime instance or None

        :param after: A datetime.datetime instance or None

        returns a list of namedtuples
        """

        # TODO: Does not search in corrections

        if before is None:
            before = datetime.now(dt.UTC)

        if after is None:
            after = FIRST_UTC_DATETIME

        if account is None:
            fk_account_pks = self._get_active_account_pks(session)
        else:
            fk_account_pks = [self._get_account_pk(session, account)]

        fk_remote_pk = None
        if jid is not None:
            fk_remote_pk = self._get_jid_pk(session, jid)

        stmt = select(Message).where(Message.fk_account_pk.in_(fk_account_pks))

        if fk_remote_pk is not None:
            stmt = stmt.where(Message.fk_remote_pk == fk_remote_pk)

        if from_users is not None:
            lowercase_users = list(map(str.lower, from_users))
            stmt = stmt.where(sa.func.lower(Message.resource).in_(lowercase_users))

        stmt = (
            stmt.where(
                Message.text.ilike(f"%{query}%"),
                Message.timestamp.between(after, before),
                ~Message.moderation.has(),
                ~Message.retraction.has(),
            )
            .order_by(sa.desc(Message.timestamp), sa.desc(Message.pk))
            .execution_options(yield_per=25)
        )

        self._explain(session, stmt)
        yield from session.scalars(stmt)

    @with_session
    @timeit
    def get_days_containing_messages(
        self, session: Session, account: str, jid: JID, year: int, month: int
    ) -> Sequence[int]:
        """
        Get days in month of year where messages for account/jid exist
        """

        fk_account_pk = self._get_account_pk(session, account)
        fk_remote_pk = self._get_jid_pk(session, jid)

        # The user wants all days which have messages within a month.
        # A message in the database with a timestamp 2024-01-01 00:00:01 UTC
        # should show up if a user is in the timezone -01:00 and passes
        # year=2023, month=12 to this function. Because for the user the message
        # happened 2023-12 localtime.
        #
        # astimezone(timezone.utc) applied to a naive datetime assumes that the
        # datetime is in localtime and converts to UTC.
        #
        # Example (Executed in a +01:00 timezone):
        #   >>> dt = datetime(2024, 1, 1).astimezone(timezone.utc)
        #   datetime.datetime(2023, 12, 31, 23, 0, tzinfo=datetime.timezone.utc)
        #
        # If we now search with this UTC timestamp in the database it will return
        # the day 31 although the user would want to see the day 1. To compensate
        # for this, the day selected is converted back by SQL to localtime.

        local_start = datetime(year, month, 1)
        _, days = calendar.monthrange(year, month)
        local_end = datetime.combine(datetime(year, month, days), dt.time.max)

        start = local_start.astimezone(dt.UTC)
        end = local_end.astimezone(dt.UTC)

        stmt = (
            select(
                sa.func.cast(
                    sa.func.strftime("%d", Message.timestamp, "unixepoch", "localtime"),
                    sa.INTEGER,
                )
            )
            .distinct()
            .where(
                Message.fk_remote_pk == fk_remote_pk,
                Message.fk_account_pk == fk_account_pk,
                Message.timestamp.between(start, end),
                Message.correction_id.is_(None),
            )
        )

        self._explain(session, stmt)
        return session.scalars(stmt).all()

    def _get_message_ts(
        self,
        session: Session,
        account: str,
        jid: JID,
        direction: Literal["first", "last"],
    ) -> datetime | None:
        fk_account_pk = self._get_account_pk(session, account)
        fk_remote_pk = self._get_jid_pk(session, jid)

        stmt = select(Message.timestamp).where(
            Message.fk_remote_pk == fk_remote_pk,
            Message.fk_account_pk == fk_account_pk,
            Message.correction_id.is_(None),
        )
        if direction == "first":
            stmt = stmt.order_by(sa.desc(Message.timestamp), sa.desc(Message.pk))
        else:
            stmt = stmt.order_by(Message.timestamp, Message.pk)

        stmt = stmt.limit(1)

        self._explain(session, stmt)
        return session.scalar(stmt)

    @with_session
    @timeit
    def get_last_message_ts(
        self, session: Session, account: str, jid: JID
    ) -> datetime | None:
        """
        Get the timestamp of the last message we received for the jid
        """
        return self._get_message_ts(session, account, jid, direction="first")

    @with_session
    @timeit
    def get_first_message_ts(
        self, session: Session, account: str, jid: JID
    ) -> datetime | None:
        """
        Get the timestamp of the first message we received for the jid
        """
        return self._get_message_ts(session, account, jid, direction="last")

    @with_session
    @timeit
    def get_first_message_meta_for_date(
        self, session: Session, account: str, jid: JID, date: dt.date
    ) -> tuple[int, datetime] | None:
        """
        Load meta data (pk, timestamp) for the first message of
        a specific date

        for details about finding start and end of a day see
        get_days_containing_messages()
        """

        fk_account_pk = self._get_account_pk(session, account)
        fk_remote_pk = self._get_jid_pk(session, jid)

        start = datetime.combine(date, dt.time.min).astimezone(dt.UTC)
        end = datetime.combine(date, dt.time.max).astimezone(dt.UTC)

        stmt = (
            select(Message.pk, Message.timestamp)
            .where(
                Message.fk_remote_pk == fk_remote_pk,
                Message.fk_account_pk == fk_account_pk,
                Message.timestamp.between(start, end),
                Message.correction_id.is_(None),
            )
            .order_by(Message.timestamp, Message.pk)
            .limit(1)
        )

        self._explain(session, stmt)
        result = session.execute(stmt).one_or_none()
        if result is None:
            return None
        return result.pk, result.timestamp

    @with_session
    @timeit
    def get_recent_muc_nicks(
        self, session: Session, account: str, jid: JID
    ) -> Sequence[str]:
        fk_account_pk = self._get_account_pk(session, account)
        fk_remote_pk = self._get_jid_pk(session, jid)

        recent = datetime.now(dt.UTC) - timedelta(days=30)

        stmt = (
            select(Message.resource)
            .distinct()
            .where(
                Message.fk_remote_pk == fk_remote_pk,
                Message.fk_account_pk == fk_account_pk,
                Message.timestamp > recent,
                Message.direction == ChatDirection.INCOMING,
                Message.resource.isnot(None),
                Message.correction_id.is_(None),
            )
            .order_by(sa.desc(Message.timestamp))
        )

        self._explain(session, stmt)
        return cast(Sequence[str], session.scalars(stmt).all())

    @with_session
    @timeit
    def get_mam_archive_state(
        self, session: Session, account: str, jid: JID
    ) -> MAMArchiveState | None:
        fk_account_pk = self._get_account_pk(session, account)
        fk_remote_pk = self._get_jid_pk(session, jid)

        stmt = select(MAMArchiveState).where(
            MAMArchiveState.fk_account_pk == fk_account_pk,
            MAMArchiveState.fk_remote_pk == fk_remote_pk,
        )

        self._explain(session, stmt)
        return session.scalar(stmt)

    @with_session
    @timeit
    def reset_mam_archive_state(self, session: Session, account: str, jid: JID) -> None:
        fk_account_pk = self._get_account_pk(session, account)
        fk_remote_pk = self._get_jid_pk(session, jid)

        stmt = delete(MAMArchiveState).where(
            MAMArchiveState.fk_account_pk == fk_account_pk,
            MAMArchiveState.fk_remote_pk == fk_remote_pk,
        )

        session.execute(stmt)

    @with_session
    @timeit
    def update_pending_message(
        self,
        session: Session,
        account: str,
        jid: JID,
        message_id: str,
        stanza_id: str | None,
    ) -> int | None:
        fk_account_pk = self._get_account_pk(session, account)
        fk_remote_pk = self._get_jid_pk(session, jid)

        stmt = (
            update(Message)
            .where(
                Message.id == message_id,
                Message.fk_remote_pk == fk_remote_pk,
                Message.fk_account_pk == fk_account_pk,
                Message.direction == ChatDirection.OUTGOING,
            )
            .values(state=MessageState.ACKNOWLEDGED, stanza_id=stanza_id)
            .returning(Message.pk)
        )

        self._explain(session, stmt)
        return session.scalar(stmt)

    @with_session
    @timeit
    def remove_history_for_jid(self, session: Session, account: str, jid: JID) -> None:
        """
        Remove messages and metadata for a specific jid.
        """

        fk_account_pk = self._get_account_pk(session, account)
        fk_remote_pk = self._get_jid_pk(session, jid)

        tables = [
            MessageError,
            Moderation,
            Retraction,
            Receipt,
            DisplayedMarker,
            Reaction,
            Message,
            # Tables with fk constraints
            Thread,
            SecurityLabel,
        ]

        for table in tables:
            stmt = delete(table).where(
                table.fk_account_pk == fk_account_pk,
                table.fk_remote_pk == fk_remote_pk,
            )

            session.execute(stmt)

        log.info("Removed history for: %s", jid)

    @with_session
    @timeit
    def remove_all_history(self, session: Session) -> None:
        """
        Remove all message related data for all accounts
        """

        session.execute(delete(MessageError))
        session.execute(delete(Moderation))
        session.execute(delete(Retraction))
        session.execute(delete(Receipt))
        session.execute(delete(DisplayedMarker))
        session.execute(delete(Reaction))
        session.execute(delete(Message))

        # Tables with fk constraints
        session.execute(delete(SecurityLabel))
        session.execute(delete(Thread))
        session.execute(delete(Encryption))

        log.info("Removed all chat history")

    @with_session
    @timeit
    def remove_account(self, session: Session, account: str) -> None:
        fk_account_pk = self._get_account_pk(session, account)

        session.execute(delete(Account).where(Account.pk == fk_account_pk))

        self._account_pks.pop(account)

    @with_session
    def remove_og(self, session: Session, pk: int) -> None:
        session.execute(delete(OpenGraph).where(OpenGraph.pk == pk))

    @with_session
    @timeit
    def cleanup_chat_history(self, session: Session) -> None:
        """
        Remove messages from account where messages are older than max_age
        """
        for account in app.settings.get_accounts():
            max_age = app.settings.get_account_setting(account, "chat_history_max_age")
            if max_age == -1:
                continue

            fk_account_pk = self._get_account_pk(session, account)

            now = datetime.now(dt.UTC)
            threshold = now - timedelta(seconds=max_age)

            stmt = (
                select(Message)
                .where(
                    Message.fk_account_pk == fk_account_pk,
                    Message.timestamp < threshold,
                )
                .options(selectinload(Message.markers))
            )

            for message in session.scalars(stmt).unique().all():
                self._delete_message(session, message)

            log.info("Removed messages older then %s", threshold.isoformat())

    @with_session_yield_from
    @timeit
    def get_messages_for_export(
        self, session: Session, account: str, jid: JID
    ) -> Iterator[Message]:
        fk_account_pk = self._get_account_pk(session, account)
        fk_remote_pk = self._get_jid_pk(session, jid)

        stmt = (
            select(Message)
            .where(
                Message.fk_account_pk == fk_account_pk,
                Message.fk_remote_pk == fk_remote_pk,
                Message.correction_id.is_(None),
                ~Message.moderation.has(),
                ~Message.retraction.has(),
            )
            .order_by(Message.timestamp, Message.pk)
            .execution_options(yield_per=25)
        )

        self._explain(session, stmt)
        yield from session.scalars(stmt)

    @with_session
    @timeit
    def get_blocked_occupants(
        self,
        session: Session,
        account: str,
        jid: JID | None = None,
    ) -> Sequence[Occupant]:
        fk_account_pk = self._get_account_pk(session, account)

        stmt = select(Occupant)

        if jid is not None:
            fk_remote_pk = self._get_jid_pk(session, jid)
            stmt = stmt.where(Occupant.fk_remote_pk == fk_remote_pk)

        stmt = stmt.where(
            Occupant.fk_account_pk == fk_account_pk,
            Occupant.blocked == True,  # noqa: E712
        ).options(joinedload(Occupant.remote))

        self._explain(session, stmt)
        return session.scalars(stmt).all()

    @with_session
    @timeit
    def set_block_occupant(
        self,
        session: Session,
        account: str,
        jid: JID | None,
        occupant_ids: Sequence[str],
        value: bool,
    ) -> int:
        fk_account_pk = self._get_account_pk(session, account)

        stmt = update(Occupant)
        if occupant_ids:
            stmt = stmt.where(Occupant.id.in_(occupant_ids))

        if jid is not None:
            fk_remote_pk = self._get_jid_pk(session, jid)
            stmt = stmt.where(Occupant.fk_remote_pk == fk_remote_pk)

        stmt = stmt.where(
            Occupant.fk_account_pk == fk_account_pk,
            Occupant.blocked != value,
        ).values(blocked=value)

        self._explain(session, stmt)

        res = cast(CursorResult[Any], session.execute(stmt))
        return res.rowcount

    @with_session
    @timeit
    def get_occupant_by_jid(
        self,
        session: Session,
        account: str,
        room_jid: JID,
        occupant_jid: JID,
        *,
        max_age: timedelta = timedelta(),
    ) -> Occupant | None:
        cache_key = (account, room_jid, occupant_jid)

        if cache_result := self._occupant_cache.get(cache_key):
            occupant, cache_dt = cache_result
            if utc_now() - cache_dt < max_age:
                log.debug("Return result from cache")
                return occupant

        fk_account_pk = self._get_account_pk(session, account)
        fk_remote_pk = self._get_jid_pk(session, room_jid)
        fk_real_remote_pk = self._get_jid_pk(session, occupant_jid)

        stmt = select(Occupant).where(
            Occupant.fk_remote_pk == fk_remote_pk,
            Occupant.fk_account_pk == fk_account_pk,
            Occupant.fk_real_remote_pk == fk_real_remote_pk,
        )

        self._explain(session, stmt)
        res = session.scalars(stmt).all()
        if not res:
            log.debug("Unable to find occupant for jid %s", occupant_jid)
            return None

        # Return latest occupant in case of multiple rows
        occupant = sorted(res, key=operator.attrgetter("updated_at"))[-1]
        self._occupant_cache[cache_key] = (occupant, utc_now())
        return occupant



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# Util module

from __future__ import annotations

from typing import Any
from typing import ParamSpec
from typing import TypeVar

import logging
from collections.abc import Callable
from functools import partial
from functools import wraps

from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Message
from nbxmpp.structs import MessageProperties
from nbxmpp.task import Task

from gajim.common import app
from gajim.common import types

T = TypeVar("T")
P = ParamSpec("P")


def from_xs_boolean(value: str | bool) -> bool:
    if isinstance(value, bool):
        return value

    if value in ('1', 'true', 'True'):
        return True

    if value in ('0', 'false', 'False', ''):
        return False

    raise ValueError(f'Cant convert {value} to python boolean')


def to_xs_boolean(value: bool | None) -> str:
    # Convert to xs:boolean ('true', 'false')
    # from a python boolean (True, False) or None
    if value is True:
        return 'true'

    if value is False:
        return 'false'

    if value is None:
        return 'false'

    raise ValueError(f'Cant convert {value} to xs:boolean')


def event_node(node: str) -> Callable[[Callable[P, object]], Callable[P, None]]:
    def event_node_decorator(func: Callable[P, object]) -> Callable[P, None]:
        @wraps(func)
        def func_wrapper(self: Any,
                         _con: types.NBXMPPClient,
                         _stanza: Message,
                         properties: MessageProperties
                         ) -> None:
            if not properties.is_pubsub_event:
                return

            assert properties.pubsub_event is not None
            if properties.pubsub_event.node != node:
                return
            func(self, _con, _stanza, properties)

        return func_wrapper
    return event_node_decorator


def store_publish(func: Callable[P, T]) -> Callable[P, T | None]:
    @wraps(func)
    def func_wrapper(self: Any, *args: P.args, **kwargs: P.kwargs) -> T | None:
        if not app.account_is_connected(self._account):
            self._stored_publish = partial(func, self, *args, **kwargs)
            return None
        return func(self, *args, **kwargs)
    return func_wrapper


class LogAdapter(logging.LoggerAdapter[logging.Logger]):
    def process(self, msg: str, kwargs: Any) -> tuple[str, Any]:
        assert self.extra is not None
        return f'({self.extra["account"]}) {msg}', kwargs


def as_task(func: Callable[P, T]) -> Callable[P, T]:
    @wraps(func)
    def func_wrapper(
        self: Any,
        *args: Any,
        timeout: int | None = None,
        callback: Callable[..., Any] | None = None,
        user_data: Any = None,
        **kwargs: Any,
    ) -> T:
        task_ = Task(func(self, *args, **kwargs))
        task_.set_timeout(timeout)
        app.register_task(self, task_)
        task_.set_finalize_func(app.remove_task, id(self))
        task_.set_user_data(user_data)
        if callback is not None:
            task_.add_done_callback(callback)
        task_.start()
        return task_
    return func_wrapper


def prepare_stanza(stanza: Message, plaintext: str) -> None:
    delete_nodes(stanza, 'encrypted', Namespace.OMEMO_TEMP)
    delete_nodes(stanza, 'body')
    stanza.setBody(plaintext)


def delete_nodes(stanza: Message,
                 name: str,
                 namespace: str | None = None
                 ) -> None:

    nodes = stanza.getTags(name, namespace=namespace)
    for node in nodes:
        stanza.delChild(node)



# Copyright (C) 2006 Gustavo J. A. M. Carneiro <gjcarneiro AT gmail.com>
#                    Nikos Kouremenos <kourem AT gmail.com>
# Copyright (C) 2006-2014 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2007 Jean-Marie Traissard <jim AT lapin.org>
#                    Julien Pivotto <roidelapluie AT gmail.com>
# Copyright (C) 2008 Stephan Erb <steve-e AT h3c.de>
# Copyright (c) 2009 Thorsten Glaser <t.glaser AT tarent.de>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import cast

import logging

import keyring

from gajim.common import app
from gajim.common.util.version import package_version

__all__ = [
    'delete_password',
    'get_password',
    'init',
    'is_keyring_available',
    'save_password'
]

log = logging.getLogger('gajim.c.passwords')


class Interface:
    def __init__(self):
        self.backend = cast(keyring.backend.KeyringBackend, None)
        self._is_keyring_available = False

    @property
    def is_keyring_available(self):
        return self._is_keyring_available

    def init(self) -> None:
        backends = keyring.backend.get_all_keyring()
        for backend in backends:
            log.info('Found keyring backend: %s', backend)

        if (app.settings.get('enable_keepassxc_integration') and
                package_version('keyring>=23.8.1')):
            _keyring = keyring.get_keyring()
            self.backend = _keyring.with_properties(scheme='KeePassXC')
        else:
            self.backend = keyring.get_keyring()
        log.info('Select %s backend', self.backend)

        self._is_keyring_available = keyring.core.recommended(self.backend)


class SecretPasswordStorage:
    '''
    Store password using Keyring
    '''

    @staticmethod
    def save_password(account_name: str, password: str) -> bool:
        if not is_keyring_available():
            log.warning('No recommended keyring backend available.'
                        'Passwords cannot be stored.')
            return False

        account_jid = app.get_jid_from_account(account_name)

        try:
            log.info('Save password to keyring')
            _interface.backend.set_password('gajim', account_jid, password)
            return True
        except Exception:
            log.exception('Save password failed')
            return False

    @staticmethod
    def get_password(account_name: str) -> str | None:
        if not is_keyring_available():
            return

        account_jid = app.get_jid_from_account(account_name)

        log.info('Request password from keyring for %s', account_jid)

        try:
            # For security reasons remove clear-text password
            ConfigPasswordStorage.delete_password(account_name)
            password = _interface.backend.get_password('gajim', account_jid)
        except Exception:
            log.exception('Request password failed')
            return

        if password is not None:
            return password

        # Migration from account name to account jid
        try:
            password = _interface.backend.get_password('gajim', account_name)
        except Exception:
            log.exception('Request password failed')
            return

        if password is not None:
            result = SecretPasswordStorage.save_password(account_name, password)
            if not result:
                log.error('Password migration failed')

        return password

    @staticmethod
    def delete_password(account_name: str) -> None:
        if not is_keyring_available():
            return

        log.info('Remove password from keyring')

        account_jid = app.get_jid_from_account(account_name)

        keyring_errors = (
            keyring.errors.InitError,
            keyring.errors.KeyringLocked,
            keyring.errors.PasswordDeleteError,
        )
        try:
            _interface.backend.delete_password('gajim', account_name)
        except keyring_errors:
            pass
        except Exception:
            pass

        try:
            return _interface.backend.delete_password('gajim', account_jid)
        except keyring_errors as error:
            log.warning('Removing password failed: %s', error)
        except Exception:
            log.exception('Removing password failed')


class ConfigPasswordStorage:
    '''
    Store password directly in Gajim's config
    '''

    @staticmethod
    def get_password(account_name: str) -> str:
        return app.settings.get_account_setting(account_name, 'password')

    @staticmethod
    def save_password(account_name: str, password: str) -> bool:
        app.settings.set_account_setting(account_name, 'password', password)
        return True

    @staticmethod
    def delete_password(account_name: str) -> None:
        app.settings.set_account_setting(account_name, 'password', '')


class MemoryPasswordStorage:
    '''
    Store password in memory
    '''

    _passwords: dict[str, str] = {}

    def get_password(self, account_name: str) -> str:
        return self._passwords.get(account_name, '')

    def save_password(self, account_name: str, password: str) -> bool:
        self._passwords[account_name] = password
        return True

    def delete_password(self, account_name: str) -> None:
        self._passwords[account_name] = ''


def init() -> None:
    _interface.init()


def is_keyring_available() -> bool:
    return _interface.is_keyring_available


def get_password(account_name: str) -> str | None:
    if not app.settings.get_account_setting(account_name, 'savepass'):
        return MemoryPasswordStorage().get_password(account_name)

    if app.settings.get('use_keyring'):
        return SecretPasswordStorage.get_password(account_name)
    return ConfigPasswordStorage.get_password(account_name)


def save_password(account_name: str, password: str) -> bool:
    if not app.settings.get_account_setting(account_name, 'savepass'):
        return MemoryPasswordStorage().save_password(account_name, password)

    if app.settings.get('use_keyring'):
        return SecretPasswordStorage.save_password(account_name, password)
    return ConfigPasswordStorage.save_password(account_name, password)


def delete_password(account_name: str) -> None:
    if app.settings.get('use_keyring'):
        return SecretPasswordStorage.delete_password(account_name)
    return ConfigPasswordStorage.delete_password(account_name)


_interface = Interface()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# All XEPs that don’t need their own module


import logging

from nbxmpp.structs import MessageProperties

from gajim.common.storage.archive import models as mod

log = logging.getLogger('gajim.c.m.misc')


# XEP-0066: Out of Band Data
def parse_oob(properties: MessageProperties) -> list[mod.OOB]:
    if not properties.is_oob:
        return []

    assert properties.oob is not None

    return [
        mod.OOB(
        url=properties.oob.url,
        description=properties.oob.desc)
    ]



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import typing
from typing import NotRequired
from typing import TypedDict

import base64
import dataclasses
from html.parser import HTMLParser

import nbxmpp.structs

from gajim.common.util.text import to_one_line

if typing.TYPE_CHECKING:
    from gajim.common.storage.archive import models as mod


MAX_ATTRIBUTE_SIZE = 200
ALLOWED_OG_PROPERTIES = [
    "og:title",
    "og:description",
    "og:image",
]


@dataclasses.dataclass
class OpenGraphThumbnail:
    uri: str
    type: str
    data: bytes

    @classmethod
    def from_bytes(cls, data: bytes, media_type: str) -> OpenGraphThumbnail:
        base64_data = base64.b64encode(data).decode()
        uri = f"data:{media_type};base64,{base64_data}"
        return cls(uri=uri, type=media_type, data=data)


@dataclasses.dataclass
class OpenGraphData:
    title: str
    description: str | None = None
    image: str | None = None
    thumbnail: OpenGraphThumbnail | None = None

    @classmethod
    def from_model(cls, row: mod.OpenGraph) -> OpenGraphData:
        thumbnail = None
        if row.image_bytes is not None:
            assert row.image_type is not None
            thumbnail = OpenGraphThumbnail.from_bytes(row.image_bytes, row.image_type)

        return cls(
            title=row.title,
            description=row.description,
            thumbnail=thumbnail,
        )

    def to_nbxmpp(self) -> nbxmpp.structs.OpenGraphData:
        attributes = dataclasses.asdict(self)
        attributes.pop("thumbnail", None)
        if self.thumbnail:
            attributes["image"] = self.thumbnail.uri
        return nbxmpp.structs.OpenGraphData(**attributes)


class OpenGraphAttributes(TypedDict):
    title: NotRequired[str]
    description: NotRequired[str]
    image: NotRequired[str]


class OpenGraphParser(HTMLParser):
    def __init__(self) -> None:
        super().__init__()

        self._attributes: OpenGraphAttributes = {}
        self._fallback_title: str = ""
        self._fallback_description: str = ""
        self._current_data = ""

    def handle_starttag(self, tag: str, attrs: list[tuple[str, str | None]]) -> None:
        if tag != "meta":
            return

        match dict(attrs):
            case {"property": "og:image", "content": content} if content:
                self._attributes["image"] = content

            case {"property": prop, "content": content} if (
                prop in ALLOWED_OG_PROPERTIES and content
            ):
                self._attributes[prop.removeprefix("og:")] = self._cleanup(
                    content[:MAX_ATTRIBUTE_SIZE]
                )

            case {"name": "description", "content": content} if content:
                self._fallback_description = self._cleanup(content[:MAX_ATTRIBUTE_SIZE])

            case _:
                pass

    def handle_endtag(self, tag: str) -> None:
        if tag == "title":
            self._fallback_title = self._cleanup(self._current_data)

        elif tag == "head":
            raise EOF

    def handle_data(self, data: str) -> None:
        self._current_data = data[:MAX_ATTRIBUTE_SIZE]

    def parse(self, text: str) -> OpenGraphData | None:
        try:
            self.feed(text)
        except EOF:
            pass

        if "title" not in self._attributes:
            if not self._fallback_title:
                return None

            self._attributes["title"] = self._fallback_title

        if "description" not in self._attributes and self._fallback_description:
            self._attributes["description"] = self._fallback_description

        if not self._attributes:
            return None

        return OpenGraphData(**self._attributes)

    def _cleanup(self, text: str) -> str:
        return to_one_line(text).strip()


class EOF(Exception):
    pass
